import abc
import enum
import math
from typing import Any, Callable, Optional, Tuple, Union

import numpy as np
import scipy.sparse as sp

from cellrank._utils._enum import DEFAULT_BACKEND, Backend_t, ModeEnum
from cellrank._utils._parallelize import parallelize
from cellrank.kernels._utils import _calculate_starts, _random_normal
from cellrank.kernels.utils._similarity import SimilarityABC

__all__ = ["Deterministic", "Stochastic", "MonteCarlo"]


class VelocityModel(ModeEnum):
    DETERMINISTIC = enum.auto()
    STOCHASTIC = enum.auto()
    MONTE_CARLO = enum.auto()


class BackwardMode(ModeEnum):
    TRANSPOSE = enum.auto()
    NEGATE = enum.auto()


class ModelABC(abc.ABC):
    """Base class for all velocity models."""

    def __init__(
        self,
        conn: sp.spmatrix,
        x: np.ndarray,
        v: np.ndarray,
        similarity: Union[
            SimilarityABC,
            Callable[[np.ndarray, np.ndarray, float], Tuple[np.ndarray, np.ndarray]],
        ],
        backward_mode: Optional[BackwardMode] = None,
        softmax_scale: float = 1.0,
        dtype: np.dtype = np.float64,
    ):
        self._conn = conn.astype(dtype, copy=False)
        self._x = x.astype(dtype, copy=False)
        self._v = v.astype(dtype, copy=False)
        self._similarity = similarity
        self._backward_mode = backward_mode
        self._softmax_scale = softmax_scale
        self._dtype = dtype

    @abc.abstractmethod
    def _compute(self, ix: int, neighs_ixs: np.ndarray, **kwargs: Any) -> Tuple[np.ndarray, np.ndarray]:
        pass

    def __call__(
        self,
        n_jobs: Optional[int] = None,
        backend: Backend_t = DEFAULT_BACKEND,
        show_progress_bar: bool = True,
        **kwargs: Any,
    ) -> Tuple[np.ndarray, np.ndarray]:
        ixs = self._ixs
        return parallelize(
            self._compute_helper,
            ixs,
            n_jobs=n_jobs,
            backend=backend,
            show_progress_bar=show_progress_bar,
            as_array=False,
            extractor=lambda data: self._reconstruct_output(np.concatenate(data, axis=-1), ixs),
            unit=self._unit,
        )(**kwargs)

    def _compute_helper(self, ixs: np.ndarray, queue=None, **kwargs) -> Tuple[np.ndarray, np.ndarray]:
        indptr, indices = self._conn.indptr, self._conn.indices
        starts = _calculate_starts(indptr, ixs)
        probs_logits = np.empty((2, starts[-1]), dtype=np.float64)

        for i, ix in enumerate(ixs):
            start, end = indptr[ix], indptr[ix + 1]
            neigh_ixs = indices[start:end]
            n_neigh = len(neigh_ixs)

            ps, ls = self._compute(ix, neigh_ixs, **kwargs)
            if np.shape(ps) != (n_neigh,):
                raise ValueError(f"Expected row of shape `{(2, n_neigh)}`, found `{np.shape(ps)}`.")

            probs_logits[0, starts[i] : starts[i] + n_neigh] = ps
            probs_logits[1, starts[i] : starts[i] + n_neigh] = ls
            if queue is not None:
                queue.put(1)

        if queue is not None:
            queue.put(None)

        return probs_logits

    def _reconstruct_output(
        self,
        data: np.ndarray,
        ixs: Optional[np.ndarray] = None,
    ) -> sp.csr_matrix:
        """Transform :class:`~numpy.ndarray` into :class:`~scipy.sparse.csr_matrix`.

        Parameters
        ----------
        data
            Array of shape ``(2, nnz)``.
        ixs
            Indices that were used to sort the data.

        Returns
        -------
        The reconstructed CSR matrix.
        """

        def reconstruct(data: np.ndarray) -> Tuple[sp.csr_matrix, sp.csr_matrix]:
            data = sp.csr_matrix(
                (
                    np.array(data, copy=True),
                    np.array(conn.indices, copy=True),
                    np.array(conn.indptr, copy=True),
                )
            )
            if aixs is not None:
                data = data[aixs]
            data.eliminate_zeros()
            return data

        if data.shape != (2, self._conn.nnz):
            raise ValueError(f"Dimension or shape mismatch: `{data.shape}`, `{(2, self._conn.nnz)}`.")

        if ixs is None:
            aixs, conn = None, self._conn
        else:
            aixs = np.argsort(ixs)
            conn = self._conn[ixs]

        probs, logits = reconstruct(data[0]), reconstruct(data[1])
        close_to_1 = np.isclose(probs.sum(1), 1.0)
        if not np.all(close_to_1):
            raise ValueError(f"Matrix is not row-stochastic, `{(~close_to_1).sum()}` do not sum to 1.")

        return probs, logits

    @property
    def _ixs(self) -> np.ndarray:
        """Order in which to process rows."""
        rng = np.random.default_rng()
        ixs = np.arange(self._conn.shape[0])
        rng.shuffle(ixs)
        return ixs

    def _uniform(self, n: int) -> Tuple[np.ndarray, np.ndarray]:
        """Uninformative probability vector when velocity is 0."""
        return np.ones(n, dtype=self._dtype) / n, np.zeros((n,), dtype=self._dtype)

    @property
    def _unit(self) -> str:
        """Progress bar unit."""
        return "cell"


class Deterministic(ModelABC):
    """Deterministic model."""

    def _compute(self, ix: int, neigh_ixs: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        W = self._x[neigh_ixs, :] - self._x[ix, :]

        if self._backward_mode not in (None, BackwardMode.NEGATE):
            return self._similarity(self._v[neigh_ixs, :], -1 * W, self._softmax_scale)

        v = self._v[ix]
        if np.all(v == 0):
            return self._uniform(len(neigh_ixs))

        if self._backward_mode == BackwardMode.NEGATE:
            v *= -1.0

        return self._similarity(v[None, :], W, self._softmax_scale)


class Stochastic(ModelABC):
    """Stochastic model.

    Parameters
    ----------
    conn
        The connectivity matrix of shape ``(n_cells, n_cells)``.
    x
        Array of shape ``(n_cells, n_features)``, such as imputed RNA expression used to calculate the displacement.
    vmean
        Array of shape ``(n_cells, n_features)`` containing the velocity mean.
    vvar
        Array of shape ``(n_cells, n_features)`` containing the velocity variance.
    dtype
        The data type to convert the arrays.
    kwargs
        Additional keyword arguments for the parent class.
    """

    def __init__(
        self,
        conn: sp.spmatrix,
        x: np.ndarray,
        vmean: np.ndarray,
        vvar: np.ndarray,
        dtype: np.dtype = np.float64,
        **kwargs: Any,
    ):
        super().__init__(conn, x, vmean, dtype=dtype, **kwargs)
        if not hasattr(self._similarity, "hessian"):
            raise AttributeError("SimilarityABC scheme doesn't have a `hessian` function.")
        self._var = vvar.astype(dtype, copy=False)

    def _compute(
        self,
        ix: int,
        nbhs_ixs: np.ndarray,
    ) -> Tuple[np.ndarray, np.ndarray]:
        v = self._v[ix]
        n_neigh, n_feat = len(nbhs_ixs), self._x.shape[1]

        if np.all(v == 0):
            return self._uniform(n_neigh)
        # compute the Hessian tensor, and turn it into a matrix that has the diagonal elements in its rows
        W = self._x[nbhs_ixs, :] - self._x[ix, :]

        H = self._similarity.hessian(v, W, self._softmax_scale)
        if H.shape == (n_neigh, n_feat):
            H_diag = H
        elif H.shape == (n_neigh, n_feat, n_feat):
            H_diag = np.array([np.diag(h) for h in H])
        else:
            raise ValueError(
                f"Expected full Hessian matrix of shape `{(n_neigh, n_feat, n_feat)}` "
                f"or its diagonal of shape `{(n_neigh, n_feat)}`, found `{H.shape}`."
            )

        # compute zero order term
        p_0, c = self._similarity(v[None, :], W, self._softmax_scale)

        # compute second order term (note that the first order term cancels)
        p_2 = 0.5 * H_diag.dot(self._var[ix])

        # combine both to give the second order Taylor approximation.
        # Can sometimes be negative because we neglected higher order terms, so force it to be non-negative
        p = np.clip(p_0 + p_2, a_min=0, a_max=1)

        nan_mask = np.isnan(p)
        p[nan_mask] = 0
        c[nan_mask] = 0

        if np.all(p == 0):
            return self._uniform(n_neigh)

        sum_ = math.fsum(p)
        if not np.isclose(sum_, 1.0, rtol=1e-12):
            p[~nan_mask] = p[~nan_mask] / sum_

        return p, c

    @property
    def _ixs(self) -> np.ndarray:
        # less recompilation for JAX
        return np.argsort((self._conn != 0).sum(1).A1)[::-1]


class MonteCarlo(ModelABC):
    """Stochastic model.

    Parameters
    ----------
    conn
        The connectivity matrix of shape ``(n_cells, n_cells)``.
    x
        Array of shape ``(n_cells, n_features)``, such as imputed RNA expression used to calculate the displacement.
    vmean
        Array of shape ``(n_cells, n_features)`` containing the velocity mean.
    vvar
        Array of shape ``(n_cells, n_features)`` containing the velocity variance.
    n_samples
        Number of samples to generate from MVN distribution using ``exp`` and ``var``.
    dtype
        The data type used to convert the arrays.
    kwargs
        Additional keyword arguments for the parent class.
    """

    def __init__(
        self,
        conn: sp.spmatrix,
        x: np.ndarray,
        vmean: np.ndarray,
        vvar: np.ndarray,
        n_samples: int = 1,
        seed: Optional[int] = None,
        dtype: np.dtype = np.float64,
        **kwargs: Any,
    ):
        super().__init__(conn, x, vmean, dtype=dtype, **kwargs)
        self._var = vvar.astype(dtype, copy=False)
        self._n_samples = n_samples
        np.random.seed(seed)  # noqa: NPY002

    def _compute(
        self,
        ix: int,
        nbhs_ixs: np.ndarray,
    ) -> Tuple[np.ndarray, np.ndarray]:
        n_neigh = len(nbhs_ixs)
        W = self._x[nbhs_ixs, :] - self._x[ix, :]

        # TODO(michalk8): don't store this?
        samples = _random_normal(self._v[ix], self._var[ix], n_samples=self._n_samples)

        probs = np.zeros((n_neigh,), dtype=np.float64)
        logits = np.zeros((n_neigh,), dtype=np.float64)
        for j in range(self._n_samples):
            ps, ls = self._similarity(np.atleast_2d(samples[j]), W, self._softmax_scale)
            probs += ps
            logits += ls

        return probs / self._n_samples, logits / self._n_samples

    @property
    def _unit(self) -> str:
        return "sample"



import types
from typing import Any, Dict, Literal, Mapping, NamedTuple, Optional, Sequence, Union

import numpy as np
import pandas as pd
import scipy.sparse as sp
from pandas.api.types import infer_dtype

from anndata import AnnData

from cellrank import logging as logg
from cellrank._utils._docs import d
from cellrank._utils._enum import DEFAULT_BACKEND, Backend_t
from cellrank._utils._key import Key
from cellrank._utils._lineage import Lineage
from cellrank._utils._linear_solver import _solve_lin_system
from cellrank._utils._utils import (
    _calculate_lineage_absorption_time_means,
    _get_cat_and_null_indices,
    _pairwise,
    _process_series,
)
from cellrank.estimators.mixins._utils import (
    BaseProtocol,
    PlotMode,
    SafeGetter,
    StatesHolder,
    logger,
    shadow,
)

__all__ = ["FateProbsMixin"]


class RecTransStates(NamedTuple):
    q: Union[np.ndarray, sp.spmatrix]
    s: Union[np.ndarray, sp.spmatrix]
    trans_indices: np.ndarray
    rec_indices: np.ndarray
    name_to_ixs: Dict[str, np.ndarray]
    term_states: pd.Series
    term_states_colors: Sequence[Any]


class FateProbsProtocol(BaseProtocol):
    _term_states: StatesHolder

    @property
    def transition_matrix(self) -> Union[np.ndarray, sp.spmatrix]:
        ...

    @property
    def terminal_states(self) -> pd.Series:
        ...

    @property
    def fate_probabilities(self) -> Optional[Lineage]:
        ...

    @property
    def absorption_times(self) -> Optional[pd.DataFrame]:
        ...

    @property
    def priming_degree(self) -> Optional[pd.Series]:
        ...

    def __len__(self) -> int:
        ...

    def _compute_fate_probabilities(
        self,
        q: Union[np.ndarray, sp.spmatrix],
        s: Union[np.ndarray, sp.spmatrix],
        trans_indices: np.ndarray,
        term_states: np.ndim,
        solver: str,
        use_petsc: bool,
        n_jobs: Optional[int],
        backend: str,
        tol: float,
        show_progress_bar: bool,
        preconditioner: str,
    ) -> np.ndarray:
        ...

    def _rec_trans_states(
        self,
        keys: Optional[Sequence[str]],
        *,
        ctx: Literal["fate_probs", "time_to_absorption"],
    ) -> RecTransStates:
        ...

    def _ensure_lineage_object(
        self,
        obj: Union[str, np.ndarray, Lineage],
        *,
        kind: Literal["macrostates", "term_states", "fate_probs"],
        backward: bool,
        **kwargs: Any,
    ) -> Lineage:
        ...

    def _write_fate_probabilities(
        self,
        fate_probs: Optional[Lineage],
    ) -> str:
        ...

    def _write_absorption_times(
        self,
        abs_times: Optional[pd.DataFrame],
        params: Mapping[str, Any] = types.MappingProxyType({}),
    ) -> str:
        ...

    def _write_lineage_priming(
        self,
        priming_degree: Optional[pd.Series],
    ) -> str:
        ...

    def _plot_continuous(
        self,
        _data: Lineage,
        _colors: Optional[np.ndarray] = None,
        _title: Optional[str] = None,
        states: Optional[Union[str, Sequence[str]]] = None,
        color: Optional[str] = None,
        mode: Literal["embedding", "time"] = PlotMode.EMBEDDING,
        time_key: Optional[str] = None,
        title: Optional[Union[str, Sequence[str]]] = None,
        same_plot: bool = True,
        cmap: str = "viridis",
        **kwargs: Any,
    ) -> None:
        ...


class FateProbsMixin:
    """Mixin that supports computation of fate probabilities and mean times to absorption."""

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)

        self._fate_probabilities: Optional[Lineage] = None
        self._absorption_times: Optional[pd.DataFrame] = None
        self._priming_degree: Optional[pd.Series] = None

    @property
    @d.get_summary(base="fate_probs")
    def fate_probabilities(self) -> Optional[Lineage]:
        """Fate probabilities.

        Informally, given a (finite, discrete) Markov chain with a set of transient states :math:`T` and
        a set of absorbing states :math:`A`, the absorption probability for cell :math:`i` from :math:`T`
        to reach cell :math:`j` from :math:`R` is the probability that a random walk initialized in :math:`i`
        will reach absorbing state :math:`j`.

        In our context, states correspond to cells, in particular, absorbing states correspond to cells
        in :attr:`terminal_states`.
        """
        return self._fate_probabilities

    @property
    @d.get_summary(base="abs_times")
    def absorption_times(self) -> Optional[pd.DataFrame]:
        """Mean and variance of the time until absorption.

        Related to conditional mean first passage times. Corresponds to the expectation of the time until absorption,
        depending on initialization, and the variance.
        """
        return self._absorption_times

    @property
    @d.get_summary(base="priming_degree")
    def priming_degree(self) -> Optional[pd.Series]:
        """Priming degree.

        Given a cell :math:`i` and a set of terminal states, this quantifies how committed vs. naive cell :math:`i` is,
        i.e. its degree of pluripotency. Low values correspond to naive cells (high degree of pluripotency), high values
        correspond to committed cells (low degree of pluripotency).
        """
        return self._priming_degree

    @d.dedent
    def compute_fate_probabilities(
        self: FateProbsProtocol,
        keys: Optional[Sequence[str]] = None,
        solver: Union[str, Literal["direct", "gmres", "lgmres", "bicgstab", "gcrotmk"]] = "gmres",
        use_petsc: bool = True,
        n_jobs: Optional[int] = None,
        backend: Backend_t = DEFAULT_BACKEND,
        show_progress_bar: bool = True,
        tol: float = 1e-6,
        preconditioner: Optional[str] = None,
    ) -> None:
        """Compute fate probabilities.

        For each cell, this computes the probability of being absorbed in any of the :attr:`terminal_states`. In
        particular, this corresponds to the probability that a random walk initialized in transient cell :math:`i`
        will reach any cell from a fixed transient state before reaching a cell from any other transient state.

        Parameters
        ----------
        keys
            Terminal states for which to compute the fate probabilities.
            If :obj:`None`, use all states defined in :attr:`terminal_states`.
        %(absorption_utils)s

        Returns
        -------
        Nothing, just updates the following fields:

        - :attr:`fate_probabilities` - %(fate_probs.summary)s
        """
        start = logg.info("Computing fate probabilities")
        data = self._rec_trans_states(keys, ctx="fate_probs")
        fate_probs = self._compute_fate_probabilities(
            data.q,
            data.s,
            trans_indices=data.trans_indices,
            term_states=data.term_states,
            solver=solver,
            use_petsc=use_petsc,
            n_jobs=n_jobs,
            backend=backend,
            tol=tol,
            show_progress_bar=show_progress_bar,
            preconditioner=preconditioner,
        )
        fate_probs = Lineage(
            fate_probs,
            names=list(data.term_states.cat.categories),
            colors=data.term_states_colors,
        )

        params = self._create_params(remove=["use_petsc", "n_jobs", "backend", "show_progress_bar"])
        self._write_fate_probabilities(
            fate_probs,
            params=params,
            time=start,
        )

    @d.dedent
    def plot_fate_probabilities(
        self: FateProbsProtocol,
        states: Optional[Union[str, Sequence[str]]] = None,
        color: Optional[str] = None,
        mode: Literal["embedding", "time"] = PlotMode.EMBEDDING,
        time_key: Optional[str] = None,
        same_plot: bool = True,
        title: Optional[Union[str, Sequence[str]]] = None,
        cmap: str = "viridis",
        **kwargs: Any,
    ) -> None:
        """Plot fate probabilities.

        Parameters
        ----------
        states
            Subset of the macrostates to show. If :obj:`None`, plot all macrostates.
        color
            Key in :attr:`~anndata.AnnData.obs` or :attr:`anndata.AnnData.var` used to color the observations.
        mode
            Whether to plot the probabilities in an embedding or along the pseudotime.
        time_key
            Key in :attr:`~anndata.AnnData.obs` where pseudotime is stored. Only used when ``mode = 'time'``.
        title
            Title of the plot.
        same_plot
            Whether to plot the data on the same plot or not. Only use when ``mode = 'embedding'``.
            If `True` and ``discrete = False``, ``color`` is ignored.
        cmap
            Colormap for continuous annotations.
        kwargs
            Keyword arguments for :func:`~scvelo.pl.scatter`.

        Returns
        -------
        %(just_plots)s
        """
        if self.fate_probabilities is None:
            raise RuntimeError("Compute fate probabilities first as `.compute_fate_probabilities()`.")

        return self._plot_continuous(
            _data=self.fate_probabilities,
            _colors=self.fate_probabilities.colors,
            _title="fate probabilities",
            states=states,
            color=color,
            mode=mode,
            time_key=time_key,
            same_plot=same_plot,
            title=title,
            cmap=cmap,
            **kwargs,
        )

    @d.dedent
    def compute_absorption_times(
        self: FateProbsProtocol,
        keys: Optional[Sequence[str]] = None,
        calculate_variance: bool = False,
        solver: Union[str, Literal["direct", "gmres", "lgmres", "bicgstab", "gcrotmk"]] = "gmres",
        use_petsc: bool = True,
        n_jobs: Optional[int] = None,
        backend: Backend_t = DEFAULT_BACKEND,
        show_progress_bar: Optional[bool] = None,
        tol: float = 1e-6,
        preconditioner: Optional[str] = None,
    ) -> None:
        """Compute the mean time to absorption and optionally its variance.

        Parameters
        ----------
        keys
            Terminal states for which to compute the fate probabilities.
            If :obj:`None`, use all states defined in :attr:`terminal_states`.
        calculate_variance
            Whether to calculate the variance.
        %(absorption_utils)s

        Returns
        -------
        Nothing, just updates the following fields:

        - :attr:`absorption_times` - %(abs_times.summary)s
        """
        start = logg.info("Computing absorption times")
        if show_progress_bar is None:
            # prevent from displaying too many progress bars
            show_progress_bar = not calculate_variance

        data = self._rec_trans_states(keys, ctx="time_to_absorption")
        abs_times = _calculate_lineage_absorption_time_means(
            data.q,
            data.s,
            calculate_variance=calculate_variance,
            trans_indices=data.trans_indices,
            ixs=data.name_to_ixs,
            solver=solver,
            use_petsc=use_petsc,
            n_jobs=n_jobs,
            backend=backend,
            tol=tol,
            show_progress_bar=show_progress_bar,
            preconditioner=preconditioner,
            index=self.adata.obs_names,
        )

        params = self._create_params(remove=["use_petsc", "n_jobs", "backend", "show_progress_bar"])
        self._write_absorption_times(
            abs_times,
            params=params,
            time=start,
        )

    @d.dedent
    def compute_lineage_priming(
        self: FateProbsProtocol,
        method: Literal["kl_divergence", "entropy"] = "kl_divergence",
        early_cells: Optional[Union[Mapping[str, Sequence[str]], Sequence[str]]] = None,
    ) -> pd.Series:
        """%(lin_pd.full_desc)s

        Parameters
        ----------
        %(lin_pd.parameters)s
            If a :class:`dict`, the key specifies a cluster key in :attr:`~anndata.AnnData.obs` and the values
            specify cluster labels containing early cells.

        Returns
        -------
        Returns the priming degree and updates the following fields:

        - :attr:`priming_degree` - %(priming_degree.summary)s
        """  # noqa: D400
        fate_probs = self.fate_probabilities
        if fate_probs is None:
            raise RuntimeError("Compute fate probabilities first as `.compute_fate_probabilities()`.")
        if isinstance(early_cells, dict):
            if len(early_cells) != 1:
                raise ValueError(f"Expected a dictionary with only 1 key, found `{len(early_cells)}`.")
            key = next(iter(early_cells.keys()))
            if key not in self.adata.obs:
                raise KeyError(f"Unable to find clusters in `adata.obs[{key!r}]`.")
            if not isinstance(self.adata.obs[key].dtype, pd.CategoricalDtype):
                raise TypeError(
                    f"Expected `adata.obs[{key!r}]` to be categorical, " f"found `{infer_dtype(self.adata.obs[key])}`."
                )
            early_cells = self.adata.obs[key].isin(early_cells[key])
        elif early_cells is not None:
            early_cells = np.asarray(early_cells)
            if not np.issubdtype(early_cells.dtype, np.bool_):
                early_cells = np.isin(self.adata.obs_names, early_cells)

        values = pd.Series(fate_probs.priming_degree(method, early_cells), index=self.adata.obs_names)
        self._write_lineage_priming(values)

        return values

    def _rec_trans_states(
        self: FateProbsProtocol,
        keys: Optional[Sequence[str]] = None,
        *,
        ctx: Literal["fate_probs", "time_to_absorption"],
    ) -> RecTransStates:
        if self.terminal_states is None:
            raise RuntimeError("Compute terminal states first as `.predict_terminal_states()`.")
        if keys is not None:
            keys = sorted(set(keys))

        # get the transition matrix
        if not sp.issparse(self.transition_matrix):
            logg.warning("Attempting to solve a potentially large linear system with dense transition matrix")

        # process the current annotations according to `keys`
        term_states, colors = _process_series(series=self.terminal_states, keys=keys, cols=self._term_states.colors)
        # warn in case only one state is left
        keys = list(term_states.cat.categories)
        if ctx == "fate_probs" and len(keys) == 1:
            logg.warning("There is only `1` terminal state, all cells will have probability `1` of going there")

        # get indices corresponding to recurrent and transient states
        rec_indices, trans_indices, name_to_ixs = _get_cat_and_null_indices(term_states)
        if not len(trans_indices):
            raise RuntimeError("Markov chain is irreducible.")

        # create Q (restriction transient-transient), S (restriction transient-recurrent)
        q = self.transition_matrix[trans_indices, :][:, trans_indices]
        s = self.transition_matrix[trans_indices, :][:, rec_indices]

        # take individual solutions and piece them together to get absorption probabilities towards the classes
        macro_ix_helper = np.cumsum([0] + [len(indices) for indices in name_to_ixs.values()])
        # `s` can be sparse or dense, ensure the correct shape
        s = np.concatenate(
            [s[:, np.arange(a, b)].sum(axis=1).reshape(-1, 1) for a, b in _pairwise(macro_ix_helper)],
            axis=1,
        )

        return RecTransStates(
            q=q,
            s=s,
            trans_indices=trans_indices,
            rec_indices=rec_indices,
            name_to_ixs=name_to_ixs,
            term_states=term_states,
            term_states_colors=colors,
        )

    def _compute_fate_probabilities(
        self: FateProbsProtocol,
        q: Union[np.ndarray, sp.spmatrix],
        s: Union[np.ndarray, sp.spmatrix],
        trans_indices: np.ndarray,
        term_states: np.ndim,
        solver: str,
        use_petsc: bool,
        n_jobs: Optional[int],
        backend: str,
        tol: float,
        show_progress_bar: bool,
        preconditioner: str,
    ) -> np.ndarray:
        _abs_classes = _solve_lin_system(
            q,
            s,
            solver=solver,
            use_petsc=use_petsc,
            n_jobs=n_jobs,
            backend=backend,
            tol=tol,
            use_eye=True,
            show_progress_bar=show_progress_bar,
            preconditioner=preconditioner,
        )
        abs_classes = np.zeros(shape=(len(self), len(term_states.cat.categories)), dtype=np.float64)
        for col, rec_class in enumerate(term_states.cat.categories):
            rec_indices = np.where(term_states == rec_class)[0]
            abs_classes[trans_indices, col] = _abs_classes[:, col]
            abs_classes[rec_indices, col] = 1.0

        mask = abs_classes >= 0
        if not np.all(mask):
            raise ValueError(
                f"`{np.sum(~mask)}` value(s) are negative. Try decreasing the tolerance as `tol=...`, "
                f"specifying a preconditioner as `preconditioner=...` or "
                f"use a direct solver as `solver='direct'` if the matrix is small."
            )
        mask = np.isclose(abs_classes.sum(1), 1.0, rtol=1e-3)
        if not np.all(mask):
            raise ValueError(
                f"`{np.sum(~mask)}` value(s) do not sum to 1 (rtol=1e-3). Try decreasing the tolerance as `tol=...`, "
                f"specifying a preconditioner as `preconditioner=...` or "
                f"use a direct solver as `solver='direct'` if the matrix is small."
            )

        return abs_classes

    @logger
    @shadow
    def _write_fate_probabilities(
        self: FateProbsProtocol,
        fate_probs: Optional[Lineage],
        params: Mapping[str, Any] = types.MappingProxyType({}),
    ) -> str:
        # fmt: off
        key = Key.obsm.fate_probs(self.backward)
        self._set("_fate_probabilities", self.adata.obsm, key=key, value=fate_probs)
        self._write_lineage_priming(None, log=False)
        self.params[key] = dict(params)
        # fmt: on

        return f"Adding `adata.obsm[{key!r}]`\n" f"       `.fate_probabilities`\n" f"    Finish"

    @logger
    @shadow
    def _write_absorption_times(
        self: FateProbsProtocol,
        abs_times: Optional[pd.DataFrame],
        params: Mapping[str, Any] = types.MappingProxyType({}),
    ) -> str:
        key = Key.obsm.abs_times(self.backward)
        self._set("_absorption_times", self.adata.obsm, key=key, value=abs_times)
        self.params[key] = dict(params)

        return f"Adding `adata.obsm[{key!r}]`\n       `.absorption_times`\n    Finish"

    @logger
    @shadow
    def _write_lineage_priming(self: FateProbsProtocol, priming_degree: Optional[pd.Series]) -> str:
        self._priming_degree = priming_degree
        key = Key.obs.priming_degree(self.backward)
        self._set("_priming_degree", self.adata.obs, key=key, value=priming_degree)

        # fmt: off
        return (
            f"Adding `adata.obs[{key!r}]`\n"
            f"       `.priming_degree`\n"
            f"    Finish"
        )
        # fmt: on

    def _read_fate_probabilities(self: FateProbsProtocol, adata: AnnData) -> bool:
        with SafeGetter(self, allowed=KeyError) as sg:
            key1 = Key.obsm.fate_probs(self.backward)
            fate_probs = self._get(
                obj=adata.obsm,
                key=key1,
                shadow_attr="obsm",
                dtype=(np.ndarray, Lineage),
            )
            self._fate_probabilities = self._ensure_lineage_object(
                fate_probs, backward=self.backward, kind="fate_probs"
            )
            key = Key.obs.priming_degree(self.backward)
            self._priming_degree = self._get(
                obj=adata.obs,
                key=key,
                shadow_attr="obs",
                dtype=pd.Series,
                allow_missing=True,
            )
            self.params[key1] = self._read_params(key1)

        return sg.ok

    def _read_absorption_times(self: FateProbsProtocol, adata: AnnData) -> bool:
        with SafeGetter(self, allowed=KeyError) as sg:
            key = Key.obsm.abs_times(self.backward)
            self._absorption_times = self._get(
                obj=adata.obsm,
                key=key,
                shadow_attr="obsm",
                dtype=pd.DataFrame,
                allow_missing=True,
            )
            self.params[key] = self._read_params(key)

        return sg.ok

    def _ensure_lineage_object(
        self: FateProbsProtocol,
        obj: Union[str, np.ndarray, Lineage],
        *,
        kind: Literal["macrostates", "term_states", "fate_probs"],
        backward: bool,
        **kwargs: Any,
    ) -> Lineage:
        if isinstance(obj, str):
            obj = getattr(self, obj)
        if isinstance(obj, Lineage):
            return obj

        try:
            return Lineage.from_adata(
                self.adata,
                backward=backward,
                kind=kind,
                estimator_backward=self.backward,
                copy=True,
                **kwargs,
            )
        except Exception as e:  # noqa: BLE001
            raise RuntimeError(f"Unable to reconstruct `.fate_probabilities`. Reason: `{e}`.") from None



from anndata import AnnData

from cellrank import logging as logg
from cellrank._utils._docs import d
from cellrank.kernels._base_kernel import UnidirectionalKernel
from cellrank.kernels.mixins import ConnectivityMixin

__all__ = ["ConnectivityKernel"]


@d.dedent
class ConnectivityKernel(ConnectivityMixin, UnidirectionalKernel):
    """Kernel which computes transition probabilities based on similarities among cells.

    As a measure of similarity, we currently support:

    - transcriptomic similarities, computed using, e.g., :func:`~scanpy.pp.neighbors`, see :cite:`wolf:18`.
    - spatial similarities, computed using, e.g., :func:`~squidpy.gr.spatial_neighbors`, see :cite:`palla:21`.

    The resulting transition matrix is symmetric and thus cannot be used to learn about the direction of the biological
    process. To include this direction, consider combining with a velocity-derived transition matrix via
    :class:`~cellrank.kernels.VelocityKernel`.

    %(density_correction)s

    Parameters
    ----------
    %(adata)s
    conn_key
        Key in :attr:`~anndata.AnnData.obsp` where connectivity matrix describing cell-cell similarity is stored.
    check_connectivity
        Check whether the underlying kNN graph is connected.
    """

    def __init__(
        self,
        adata: AnnData,
        conn_key: str = "connectivities",
        check_connectivity: bool = False,
    ):
        super().__init__(
            adata,
            conn_key=conn_key,
            check_connectivity=check_connectivity,
        )

    def compute_transition_matrix(self, density_normalize: bool = True) -> "ConnectivityKernel":
        """Compute transition matrix based on transcriptomic similarity.

        Uses symmetric, weighted kNN graph to compute symmetric transition matrix. The connectivities are computed
        using :func:`~scanpy.pp.neighbors`. Depending on the parameters used there, they can be UMAP connectivities or
        Gaussian-kernel-based connectivities with adaptive kernel width.

        Parameters
        ----------
        density_normalize
            Whether to use the underlying kNN graph for density normalization.

        Returns
        -------
        Returns self and updates :attr:`transition_matrix` and :attr:`params`.
        """
        # fmt: off
        start = logg.info(f"Computing transition matrix based on `adata.obsp[{self._conn_key!r}]`")
        if self._reuse_cache({"dnorm": density_normalize, "key": self._conn_key}, time=start):
            return self

        conn = self.connectivities
        if density_normalize:
            conn = self._density_normalize(conn)

        self.transition_matrix = conn
        logg.info("    Finish", time=start)
        # fmt: on

        return self



import dataclasses
from typing import Any, List, Mapping, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd
import scipy.sparse as sp
import scipy.stats as st
from pandas.api.types import infer_dtype
from scipy.interpolate import interp1d
from statsmodels.nonparametric.smoothers_lowess import lowess

import matplotlib.pyplot as plt
from matplotlib.collections import PolyCollection
from matplotlib.colors import to_rgb

from anndata import AnnData

from cellrank import logging as logg
from cellrank._utils._colors import _create_categorical_colors
from cellrank._utils._docs import d
from cellrank._utils._utils import _unique_order_preserving
from cellrank.kernels._utils import _ensure_numeric_ordered

__all__ = ["FlowPlotter"]

Numeric_t = Union[float, int]


@dataclasses.dataclass(frozen=True)
class Point:
    x: float
    xt: float


@d.dedent
class FlowPlotter:
    """Class that plots outgoing flow for a specific cluster :cite:`mittnenzweig:21`.

    It should be able to recreate (to a high degree) figures such as Fig. 4a in the above-mentioned paper.

    Parameters
    ----------
    %(adata)s
    tmat
        Matrix of shape ``(adata.n_obs, adata.n_obs)``.
    cluster_key
        Key in :attr:`~anndata.AnnData.obs` where clustering is stored.
    time_key
        Key in :attr:`~anndata.AnnData.obs` where experimental time is stored.
    """

    TIME_KEY = "time"

    def __init__(
        self,
        adata: AnnData,
        tmat: Union[np.ndarray, sp.spmatrix],
        cluster_key: str,
        time_key: str,
    ):
        self._adata = adata
        self._tmat = tmat
        self._ckey = cluster_key
        self._tkey = time_key

        self._cluster: Optional[str] = None
        self._clusters: Optional[Sequence[Any]] = None

        self._flow: Optional[pd.DataFrame] = None
        self._cmat: Optional[pd.DataFrame] = None

        if self._ckey not in self._adata.obs:
            raise KeyError(f"Unable to find clusters in `adata.obs[{self._ckey!r}]`.")
        if not isinstance(self._adata.obs[self._ckey].dtype, pd.CategoricalDtype):
            raise TypeError(
                f"Expected `adata.obs[{self._ckey!r}]` to be categorical, "
                f"found `{infer_dtype(self._adata.obs[self._ckey])}`."
            )
        self._adata.obs[self._tkey] = _ensure_numeric_ordered(self._adata, self._tkey)

    def prepare(
        self,
        cluster: str,
        clusters: Optional[Sequence[Any]] = None,
        time_points: Optional[Sequence[Numeric_t]] = None,
    ) -> "FlowPlotter":
        """Prepare itself for plotting by computing flow and contingency matrix.

        Parameters
        ----------
        cluster
            Source cluster for flow calculation.
        clusters
            Target clusters for flow calculation. If :obj:`None`, use all clusters.
        time_points
            Restrict flow calculation only to these time points. If :obj:`None`, use all time points.

        Returns
        -------
        Returns and modifies self.
        """
        if clusters is None:
            self._clusters = self.clusters.cat.categories
        else:
            clusters = _unique_order_preserving([cluster] + list(clusters))
            mask = self.clusters.isin(clusters).values

            self._adata = self._adata[mask]
            if not self._adata.n_obs:
                raise ValueError("No valid clusters have been selected.")
            self._tmat = self._tmat[mask, :][:, mask]
            self._clusters = [c for c in clusters if c in self.clusters.cat.categories]

        if cluster not in self._clusters:
            raise ValueError(f"Invalid source cluster `{cluster!r}`.")

        if len(self._clusters) < 2:
            raise ValueError(f"Expected at least `2` clusters, found `{len(clusters)}`.")

        if time_points is not None:
            time_points = _unique_order_preserving(time_points)
            if len(time_points) < 2:
                raise ValueError(f"Expected at least `2` time points, found `{len(time_points)}`.")

            mask = self.time.isin(time_points)

            self._adata = self._adata[mask]
            if not self._adata.n_obs:
                raise ValueError("No valid time points have been selected.")
            self._tmat = self._tmat[mask, :][:, mask]

        time_points = list(zip(self.time.cat.categories[:-1], self.time.cat.categories[1:]))

        logg.info(
            f"Computing flow from `{cluster}` into `{len(self._clusters) - 1}` cluster(s) "
            f"in `{len(time_points)}` time points"
        )
        self._cluster = cluster
        self._cmat = self.compute_contingency_matrix()
        self._flow = self.compute_flow(time_points, cluster)

        return self

    def compute_flow(
        self,
        time_points: Sequence[Tuple[Numeric_t, Numeric_t]],
        cluster: Optional[str] = None,
    ) -> pd.DataFrame:
        """Compute outgoing flow.

        Parameters
        ----------
        time_points
            Time point pair for which to calculate the flow.
        cluster
            Cluster for which to calculate the outgoing flow. If :obj:`None`, calculate the flow for all clusters.

        Returns
        -------
        Dataframe of shape ``(n_time_points, n_clusters)`` if ``cluster != None`` or
        a dataframe of shape ``(n_time_points * n_clusters, n_clusters)`` otherwise.
        The dataframe's index is a multi-index and the 1st level corresponds to time and
        the 2nd level to source clusters.
        """

        def default_helper(t1: Numeric_t, t2: Numeric_t) -> pd.DataFrame:
            subset, row_cls, col_cls = self._get_time_subset(t1, t2)

            df = pd.DataFrame(subset.A if sp.issparse(subset) else subset)
            df = df.groupby(row_cls).sum().T.groupby(col_cls).sum().T

            res = pd.DataFrame(np.zeros((n, n)), index=categories, columns=categories)
            res.loc[df.index, df.columns] = df
            res.fillna(0, inplace=True)

            return res

        def cluster_helper(t1: Numeric_t, t2: Numeric_t) -> pd.DataFrame:
            subset, row_cls, col_cls = self._get_time_subset(t1, t2, cluster=cluster)

            df = pd.DataFrame(subset.A if sp.issparse(subset) else subset).sum(0)
            df = df.groupby(col_cls).sum()
            df = pd.DataFrame([df], index=[cluster], columns=df.index)

            res = pd.DataFrame(np.zeros((1, n)), index=[cluster], columns=categories)
            res.loc[df.index, df.columns] = df
            res.fillna(0, inplace=True)

            return res

        categories = self.clusters.cat.categories
        n = len(categories)
        callback = cluster_helper if cluster is not None else default_helper
        flows, times = [], []

        for t1, t2 in time_points:
            flow = callback(t1, t2)
            times.extend([t1] * len(flow))
            flows.append(flow)

        flow = pd.concat(flows)
        flow.set_index([times, flow.index], inplace=True)
        flow /= flow.sum(1).values[:, None]
        flow.fillna(0, inplace=True)

        return flow

    def compute_contingency_matrix(self) -> pd.DataFrame:
        """Row-normalized contingency matrix of shape ``(n_clusters, n_time_points)``."""
        cmat = pd.crosstab(self.clusters, self.time)
        return (cmat / cmat.sum(0).values[None, :]).fillna(0)

    @d.get_sections(base="flow", sections=["Parameters"])
    def plot(
        self,
        min_flow: float = 0,
        remove_empty_clusters: bool = True,
        ascending: Optional[bool] = False,
        alpha: float = 0.8,
        xticks_step_size: Optional[int] = 1,
        legend_loc: Optional[str] = "upper right out",
        figsize: Optional[Tuple[float, float]] = None,
        dpi: Optional[int] = None,
    ) -> plt.Axes:
        """Plot outgoing flow.

        Parameters
        ----------
        min_flow
            Only show flow edges with flow greater than this value. Flow values are always in :math:`[0, 1]`.
        remove_empty_clusters
            Whether to remove clusters with no incoming flow edges.
        ascending
            Whether to sort the cluster by ascending or descending incoming flow.
            If `None`, use the order as in defined by ``clusters``.
        alpha
            Alpha value for cell proportions.
        xticks_step_size
            Show only every other *n-th* tick on the x-axis. If :obj:`None`, don't show any ticks.
        legend_loc
            Position of the legend. If :obj:`None`, do not show the legend.
        figsize
            Size of the figure.
        dpi
            Dots per inch.

        Returns
        -------
        The axes object.
        """
        if self._flow is None or self._cmat is None:
            raise RuntimeError("Compute flow and contingency matrix first as `.prepare()`.")

        flow, cmat = self._flow, self._cmat
        try:
            if remove_empty_clusters:
                self._remove_min_clusters(min_flow)
            logg.info(
                f"Plotting flow from `{self._cluster}` into `{len(self._flow.columns) - 1}` cluster(s) "
                f"in `{len(self._cmat.columns) - 1}` time points"
            )
            return self._plot(
                self._rename_times(),
                ascending=ascending,
                min_flow=min_flow,
                alpha=alpha,
                xticks_step_size=xticks_step_size,
                legend_loc=legend_loc,
                figsize=figsize,
                dpi=dpi,
            )
        finally:
            self._flow = flow
            self._cmat = cmat

    def _get_time_subset(
        self, t1: Numeric_t, t2: Numeric_t, cluster: Optional[str] = None
    ) -> Tuple[Union[np.ndarray, sp.spmatrix], pd.Series, pd.Series]:
        if cluster is None:
            row_ixs = np.where(self.time == t1)[0]
        else:
            row_ixs = np.where((self.time == t1) & (self.clusters == cluster))[0]

        col_ixs = np.where(self.time == t2)[0]
        row_cls = self.clusters.values[row_ixs]
        col_cls = self.clusters.values[col_ixs]

        return self._tmat[row_ixs, :][:, col_ixs], row_cls, col_cls

    def _remove_min_clusters(self, min_flow: float) -> None:
        logg.debug("Removing clusters with no incoming flow edges")
        columns = (self._flow.loc[(slice(None), self._cluster), :] > min_flow).any()
        columns = columns[columns].index
        if not len(columns):
            raise ValueError("After removing clusters with no incoming flow edges, none remain.")
        self._flow = self._flow[columns]

    def _rename_times(self) -> Sequence[Numeric_t]:
        # make sure we have enough horizontal space to draw the flow (i.e. time points are at least 1 unit apart)
        old_times = self._cmat.columns
        tmp = np.array(old_times)
        tmp = (tmp - tmp.min()) / (tmp.max() - tmp.min())
        tmp /= np.min(tmp[1:] - tmp[:-1])
        time_mapper = dict(zip(old_times, tmp))
        self._flow.index = pd.MultiIndex.from_tuples([(time_mapper[t], c) for t, c in self._flow.index])
        self._cmat.columns = tmp
        return old_times

    def _order_clusters(self, cluster: str, ascending: Optional[bool] = False) -> Tuple[List[Any], List[Any]]:
        if ascending is not None:
            tmp = [[], []]
            total_flow = self._flow.loc[(slice(None), cluster), :].sum().sort_values(ascending=ascending)
            for i, c in enumerate(c for c in total_flow.index if c != cluster):
                tmp[i % 2].append(c)
            return tmp[0][::-1], tmp[1]

        clusters = [c for c in self._clusters if c != cluster]
        return clusters[: len(clusters) // 2], clusters[len(clusters) // 2 :]

    def _calculate_y_offsets(self, clusters: Sequence[Any], delta: float = 0.2) -> Mapping[Any, float]:
        offset = [0]
        for i in range(1, len(clusters)):
            offset.append(offset[-1] + delta + np.max(self._cmat.loc[clusters[i]] + self._cmat.loc[clusters[i - 1]]))
        return dict(zip(clusters, offset))

    def _plot_smoothed_proportion(
        self,
        ax: plt.Axes,
        clusters: Sequence[Any],
        y_offset: Mapping[Any, float],
        alpha: float = 0.8,
    ) -> Tuple[Mapping[Any, np.ndarray], Mapping[Any, PolyCollection]]:
        start_t, end_t = self._cmat.columns.min(), self._cmat.columns.max()
        x = np.array(self._cmat.columns)  # fitting
        # extrapolation
        e = np.linspace(start_t, end_t, int(1 + (end_t - start_t) * 100))

        smoothed_proportion, handles = {}, {}
        for clust in clusters:
            y = self._cmat.loc[clust]
            f = interp1d(x, y)
            fe = f(e)
            lo = lowess(fe, e, frac=0.3, is_sorted=True, return_sorted=False)
            smoothed_proportion[clust] = lo

            handles[clust] = ax.fill_between(
                e,
                y_offset[clust] + lo,
                y_offset[clust] - lo,
                color=self.cmap[clust],
                label=clust,
                alpha=alpha,
                edgecolor=None,
            )

        return smoothed_proportion, handles

    def _draw_flow_edge(
        self,
        ax,
        x1: Point,
        x2: Point,
        y1: Point,
        y2: Point,
        start_color: Tuple[float, float, float],
        end_color: Tuple[float, float, float],
        flow: float,
        alpha: float = 0.8,
    ) -> None:
        # transcribed from: https://github.com/tanaylab/embflow/blob/main/scripts/generate_paper_figures/plot_vein.r
        dx = x2.xt - x1.x
        dy = y2.xt - y1.x
        dxt = x2.x - x1.x
        dyt = y2.x - y1.xt

        start_color = np.asarray(to_rgb(start_color))
        end_color = np.asarray(to_rgb(end_color))
        delta = 0.05

        beta0 = _lcdf(0)
        beta_f = _lcdf(1) - _lcdf(0)

        rs = np.arange(0, 1, delta)
        beta = (_lcdf(rs) - beta0) / beta_f
        beta5 = (_lcdf(rs + delta) - beta0) / beta_f

        sx1 = x1.x + rs * dx
        sy1 = y1.x + beta * dy
        sx2 = x1.x + (rs + delta) * dx
        sy2 = y1.x + beta5 * dy

        sx1t = x1.x + flow + rs * dxt
        sy1t = y1.xt + beta * dyt
        sx2t = x1.x + flow + (rs + delta) * dxt
        sy2t = y1.xt + beta5 * dyt

        xs = np.c_[sx1, sx2, sx2t, sx1t]
        ys = np.c_[sy1, sy2, sy2t, sy1t]

        start_alpha, end_alpha = 0.2, alpha
        if start_alpha > end_alpha:
            start_alpha, end_alpha = end_alpha, start_alpha
        col = np.c_[
            (start_color * (1 - rs[:, None])) + (end_color * rs[:, None]),
            np.linspace(start_alpha, end_alpha, len(rs)),
        ]

        for x, y, c in zip(xs, ys, col):
            ax.fill(x, y, c=c, edgecolor=None)

    def _plot(
        self,
        old_times: Sequence[Numeric_t],
        ascending: Optional[bool],
        min_flow: float = 0,
        alpha: float = 0.8,
        xticks_step_size: Optional[int] = 1,
        legend_loc: Optional[str] = "upper right out",
        figsize: Optional[Tuple[float, float]] = None,
        dpi: Optional[int] = None,
    ) -> plt.Axes:
        from cellrank.pl._utils import _position_legend

        def r(num: float) -> int:
            return max(0, int(round(num, 2) * 100) - 1)

        def draw_edges(
            curr_t: Numeric_t,
            next_t: Numeric_t,
            clusters: Sequence[Any],
            *,
            bottom: bool,
        ):
            smooth_cluster = float(smoothed_proportions[self._cluster][r(curr_t)])
            flow = self._flow.loc[curr_t]
            for clust in clusters:
                fl = flow.loc[self._cluster, clust]
                if fl > min_flow:
                    fl = np.clip(fl, 0, 0.95)
                    smooth_cluster_fl = smoothed_proportions[self._cluster][r(curr_t + fl)]

                    if bottom:
                        self._draw_flow_edge(
                            ax,
                            x1=Point(curr_t, 0),
                            x2=Point(next_t - fl, next_t - fl - 0.05),
                            y1=Point(
                                cluster_offset - smooth_cluster,
                                cluster_offset - smooth_cluster_fl,
                            ),
                            y2=Point(
                                y_offset[clust] + smoothed_proportions[clust][r(next_t)],
                                y_offset[clust] + smoothed_proportions[clust][r(next_t - fl - 0.05)],
                            ),
                            flow=fl,
                            start_color=self.cmap[self._cluster],
                            end_color=self.cmap[clust],
                            alpha=alpha,
                        )
                    else:
                        self._draw_flow_edge(
                            ax,
                            x1=Point(curr_t + fl, 0),
                            x2=Point(next_t - 0.05, next_t),
                            y1=Point(
                                cluster_offset + smooth_cluster_fl,
                                cluster_offset + smooth_cluster,
                            ),
                            y2=Point(
                                y_offset[clust] - smoothed_proportions[clust][r(next_t - fl - 0.05)],
                                y_offset[clust] - smoothed_proportions[clust][r(next_t)],
                            ),
                            flow=-fl,
                            start_color=self.cmap[self._cluster],
                            end_color=self.cmap[clust],
                            alpha=alpha,
                        )

        if xticks_step_size is not None:
            xticks_step_size = max(1, xticks_step_size)
        times = self._cmat.columns
        fig, ax = plt.subplots(figsize=figsize, dpi=dpi)

        clusters_bottom, clusters_top = self._order_clusters(self._cluster, ascending)
        all_clusters = clusters_bottom + [self._cluster] + clusters_top

        y_offset = self._calculate_y_offsets(all_clusters)
        cluster_offset = y_offset[self._cluster]

        smoothed_proportions, handles = self._plot_smoothed_proportion(ax, all_clusters, y_offset, alpha=alpha)

        for curr_t, next_t in zip(times[:-1], times[1:]):
            draw_edges(curr_t, next_t, clusters_bottom, bottom=True)
            draw_edges(curr_t, next_t, clusters_top, bottom=False)

        ax.margins(0.025)
        ax.set_title(self._cluster)
        ax.set_xlabel(self._tkey)
        ax.set_ylabel(self._ckey)
        if xticks_step_size is None:
            ax.set_xticks([])
        else:
            ax.set_xticks(times[::xticks_step_size])
            ax.set_xticklabels(old_times[::xticks_step_size])
        ax.set_yticks([])

        if legend_loc not in (None, "none"):
            _position_legend(
                ax,
                legend_loc=legend_loc,
                handles=[handles[c] for c in all_clusters[::-1]],
            )

        return ax

    @property
    def clusters(self) -> pd.Series:
        """Clusters."""
        return self._adata.obs[self._ckey]

    @property
    def time(self) -> pd.Series:
        """Time points."""
        return self._adata.obs[self._tkey]

    @property
    def cmap(self) -> Mapping[str, Any]:
        """Colormap for :attr:`clusters`."""
        return dict(
            zip(
                self.clusters.cat.categories,
                self._adata.uns.get(
                    f"{self._ckey}_colors",
                    _create_categorical_colors(len(self.clusters.cat.categories)),
                ),
            )
        )


def _lcdf(x: Union[int, float, np.ndarray], loc: float = 0.5, scale: float = 0.2) -> float:
    return st.logistic.cdf(x, loc=loc, scale=scale)



import abc
import copy
import pathlib
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union

import numpy as np
import scipy.sparse as sp

import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap

from anndata import AnnData

from cellrank import logging as logg
from cellrank._utils._docs import d, inject_docs
from cellrank._utils._utils import _normalize, _read_graph_data, save_fig
from cellrank.kernels._utils import require_tmat
from cellrank.kernels.mixins import BidirectionalMixin, IOMixin, UnidirectionalMixin
from cellrank.kernels.utils import FlowPlotter, RandomWalk, TmatProjection
from cellrank.kernels.utils._random_walk import Indices_t

__all__ = ["Kernel", "UnidirectionalKernel", "BidirectionalKernel"]

Tmat_t = Union[np.ndarray, sp.spmatrix]


class KernelExpression(IOMixin, abc.ABC):
    def __init__(
        self,
        parent: Optional["KernelExpression"] = None,
        **kwargs: Any,
    ):
        super().__init__()
        self._parent = parent
        self._normalize = parent is None
        self._transition_matrix = None
        self._params: Dict[str, Any] = {}
        self._init_kwargs = kwargs  # for `_read_from_adata`

    def __init_subclass__(cls, **_: Any) -> None:
        super().__init_subclass__()

    @abc.abstractmethod
    def compute_transition_matrix(self, *args: Any, **kwargs: Any) -> "KernelExpression":
        """Compute transition matrix.

        Parameters
        ----------
        args
            Positional arguments.
        kwargs
            Keyword arguments.

        Returns
        -------
        Modifies :attr:`transition_matrix` and returns self.
        """

    @abc.abstractmethod
    def copy(self, *, deep: bool = False) -> "KernelExpression":
        """Return a copy of self. The :attr:`adata` object is not copied."""

    @property
    @abc.abstractmethod
    def adata(self) -> AnnData:
        """Annotated data object."""

    @adata.setter
    @abc.abstractmethod
    def adata(self, value: Optional[AnnData]) -> None:
        pass

    @property
    @abc.abstractmethod
    def kernels(self) -> Tuple["KernelExpression", ...]:
        """Underlying base kernels."""

    @property
    @abc.abstractmethod
    def shape(self) -> Tuple[int, int]:
        """``(n_cells, n_cells)``."""

    @property
    @abc.abstractmethod
    def backward(self) -> Optional[bool]:
        """Direction of the process."""

    @abc.abstractmethod
    def __getitem__(self, ix: int) -> "KernelExpression":
        pass

    @abc.abstractmethod
    def __len__(self) -> int:
        pass

    @d.get_full_description(base="plot_single_flow")
    @d.get_sections(base="plot_single_flow", sections=["Parameters", "Returns"])
    @d.dedent
    @require_tmat
    def plot_single_flow(
        self,
        cluster: str,
        cluster_key: str,
        time_key: str,
        clusters: Optional[Sequence[Any]] = None,
        time_points: Optional[Sequence[Union[int, float]]] = None,
        min_flow: float = 0,
        remove_empty_clusters: bool = True,
        ascending: Optional[bool] = False,
        legend_loc: Optional[str] = "upper right out",
        alpha: Optional[float] = 0.8,
        xticks_step_size: Optional[int] = 1,
        figsize: Optional[Tuple[float, float]] = None,
        dpi: Optional[int] = None,
        save: Optional[Union[str, pathlib.Path]] = None,
        show: bool = True,
    ) -> Optional[plt.Axes]:
        """Visualize outgoing flow from a cluster of cells :cite:`mittnenzweig:21`.

        Parameters
        ----------
        cluster
            Cluster for which to visualize outgoing flow.
        cluster_key
            Key in :attr:`~anndata.AnnData.obs` where clustering is stored.
        time_key
            Key in :attr:`~anndata.AnnData.obs` where experimental time is stored.
        clusters
            Visualize flow only for these clusters. If :obj:`None`, use all clusters.
        time_points
            Visualize flow only for these time points. If :obj:`None`, use all time points.
        %(flow.parameters)s
        %(plotting)s
        show
            If :obj:`False`, return :class:`~matplotlib.axes.Axes`.

        Returns
        -------
        The axes object, if ``show = False``.
        %(just_plots)s

        Notes
        -----
        This function is a Python re-implementation of the following
        `original R function <https://github.com/tanaylab/embflow/blob/main/scripts/generate_paper_figures/plot_vein.r>`_
        with some minor stylistic differences.
        This function will not recreate the results from :cite:`mittnenzweig:21`, because there, the *Metacell* model
        :cite:`baran:19` was used to compute the flow, whereas here the transition matrix is used.
        """  # noqa: E501
        fp = FlowPlotter(self.adata, self.transition_matrix, cluster_key, time_key)
        fp = fp.prepare(cluster, clusters, time_points)

        ax = fp.plot(
            min_flow=min_flow,
            remove_empty_clusters=remove_empty_clusters,
            ascending=ascending,
            alpha=alpha,
            xticks_step_size=xticks_step_size,
            legend_loc=legend_loc,
            figsize=figsize,
            dpi=dpi,
        )

        if save is not None:
            save_fig(ax.figure, save)

        if not show:
            return ax

    @d.dedent
    @require_tmat
    def plot_random_walks(
        self,
        n_sims: int = 100,
        max_iter: Union[int, float] = 0.25,
        seed: Optional[int] = None,
        successive_hits: int = 0,
        start_ixs: Indices_t = None,
        stop_ixs: Indices_t = None,
        basis: str = "umap",
        cmap: Union[str, LinearSegmentedColormap] = "gnuplot",
        linewidth: float = 1.0,
        linealpha: float = 0.3,
        ixs_legend_loc: Optional[str] = None,
        n_jobs: Optional[int] = None,
        backend: str = "loky",
        show_progress_bar: bool = True,
        figsize: Optional[Tuple[float, float]] = None,
        dpi: Optional[int] = None,
        save: Optional[Union[str, pathlib.Path]] = None,
        **kwargs: Any,
    ) -> None:
        """Plot random walks in an embedding.

        This method simulates random walks on the Markov chain defined though the corresponding transition matrix. The
        method is intended to give qualitative rather than quantitative insights into the transition matrix. Random
        walks are simulated by iteratively choosing the next cell based on the current cell's transition probabilities.

        Parameters
        ----------
        n_sims
            Number of random walks to simulate.
        %(rw_sim.parameters)s
        start_ixs
            Cells from which to sample the starting points. If :obj:`None`, use all cells.
            %(rw_ixs)s
            For example ``{'dpt_pseudotime': [0, 0.1]}`` means that starting points for random walks
            will be sampled uniformly from cells whose pseudotime is in :math:`[0, 0.1]`.
        stop_ixs
            Cells which when hit, the random walk is terminated. If :obj:`None`, terminate after ``max_iters``.
            %(rw_ixs)s
            For example ``{'clusters': ['Alpha', 'Beta']}`` and ``successive_hits = 3`` means that the random walk will
            stop prematurely after cells in the above specified clusters have been visited successively 3 times in
            a row.
        basis
            Basis in :attr:`~anndata.AnnData.obsm` to use as an embedding.
        cmap
            Colormap for the random walk lines.
        linewidth
            Width of the random walk lines.
        linealpha
            Alpha value of the random walk lines.
        ixs_legend_loc
            Legend location for the start/top indices.
        %(parallel)s
        %(plotting)s
        kwargs
            Keyword arguments for :func:`~scvelo.pl.scatter`.

        Returns
        -------
        %(just_plots)s
        For each random walk, the first/last cell is marked by the start/end colors of ``cmap``.
        """
        rw = RandomWalk(self.adata, self.transition_matrix, start_ixs=start_ixs, stop_ixs=stop_ixs)
        sims = rw.simulate_many(
            n_sims=n_sims,
            max_iter=max_iter,
            seed=seed,
            n_jobs=n_jobs,
            backend=backend,
            successive_hits=successive_hits,
            show_progress_bar=show_progress_bar,
        )

        rw.plot(
            sims,
            basis=basis,
            cmap=cmap,
            linewidth=linewidth,
            linealpha=linealpha,
            ixs_legend_loc=ixs_legend_loc,
            figsize=figsize,
            dpi=dpi,
            save=save,
            **kwargs,
        )

    @require_tmat
    def plot_projection(
        self,
        basis: str = "umap",
        key_added: Optional[str] = None,
        recompute: bool = False,
        stream: bool = True,
        connectivities: Optional[sp.spmatrix] = None,
        **kwargs: Any,
    ) -> None:
        """Plot :attr:`transition_matrix` as a stream or a grid plot.

        Parameters
        ----------
        basis
            Key in :attr:`~anndata.AnnData.obsm` containing the basis.
        key_added
            If not :obj:`None`, save the result to :attr:`adata.obsm['{key_added}'] <anndata.AnnData.obsm>`.
            Otherwise, save the result to ``'T_fwd_{basis}'`` or ``'T_bwd_{basis}'``, depending on the direction.
        recompute
            Whether to recompute the projection if it already exists.
        stream
            If :obj:`True`, use :func:`~scvelo.pl.velocity_embedding_stream`.
            Otherwise, use :func:`~scvelo.pl.velocity_embedding_grid`.
        connectivities
            Connectivity matrix to use for projection. If :obj:`None`, use ones from the underlying kernel, is possible.
        kwargs
            Keyword argument for the above-mentioned plotting function.

        Returns
        -------
        Nothing, just plots and modifies :attr:`~anndata.AnnData.obsm` with a key based on the ``key_added``.
        """
        proj = TmatProjection(self, basis=basis)
        proj.project(key_added=key_added, recompute=recompute, connectivities=connectivities)
        proj.plot(stream=stream, **kwargs)

    def __add__(self, other: "KernelExpression") -> "KernelExpression":
        return self.__radd__(other)

    def __radd__(self, other: "KernelExpression") -> "KernelExpression":
        def same_level_add(k1: "KernelExpression", k2: "KernelExpression") -> bool:
            if not (isinstance(k1, KernelAdd) and isinstance(k2, KernelMul)):
                return False

            for kexpr in k1:
                if not isinstance(kexpr, KernelMul):
                    return False
                if not kexpr._bin_consts:
                    return False
            return True

        if not isinstance(other, KernelExpression):
            return NotImplemented

        s = self * 1.0 if isinstance(self, Kernel) else self
        o = other * 1.0 if isinstance(other, Kernel) else other

        # (c1 * x + c2 * y + ...) + (c3 * z) => (c1 * x + c2 * y + c3 + ... + * z)
        if same_level_add(s, o):
            return KernelAdd(*tuple(s) + (o,))
        if same_level_add(o, s):
            return KernelAdd(*tuple(o) + (s,))

        # add virtual constant
        if not isinstance(s, KernelMul):
            s = s * 1.0
        if not isinstance(o, KernelMul):
            o = o * 1.0

        return KernelAdd(s, o)

    def __mul__(self, other: Union[float, int, "KernelExpression"]) -> "KernelExpression":
        return self.__rmul__(other)

    def __rmul__(self, other: Union[int, float, "KernelExpression"]) -> "KernelExpression":
        def same_level_mul(k1: "KernelExpression", k2: "KernelExpression") -> bool:
            return isinstance(k1, KernelMul) and isinstance(k2, Constant) and k1._bin_consts

        if isinstance(other, (int, float, np.integer, np.floating)):
            other = Constant(self.adata, other)

        if not isinstance(other, KernelExpression):
            return NotImplemented

        # fmt: off
        s = self if isinstance(self, (KernelMul, Constant)) else KernelMul(Constant(self.adata, 1.0), self)
        o = other if isinstance(other, (KernelMul, Constant)) else KernelMul(Constant(other.adata, 1.0), other)
        # fmt: on

        # at this point, only KernelMul and Constant is possible (two constants are not possible)
        # (c1 * k) * c2 => (c1 * c2) * k
        if same_level_mul(s, o):
            c, expr = s._split_const
            return KernelMul(c * o, expr)
        if same_level_mul(o, s):
            c, expr = o._split_const
            return KernelMul(c * s, expr)

        return KernelMul(s, o)

    @d.get_sections(base="write_to_adata", sections=["Parameters"])
    @inject_docs()  # gets rid of {{}} in %(write_to_adata)s
    @d.dedent
    @require_tmat
    def write_to_adata(self, key: Optional[str] = None, copy: bool = False) -> None:
        """Write the transition matrix and parameters used for computation to the underlying :attr:`adata` object.

        Parameters
        ----------
        key
            Key used when writing transition matrix to :attr:`adata`.
            If :obj:`None`, the key will be determined automatically.
        copy
            Whether to copy the :attr:`transition_matrix`.

        Returns
        -------
        %(write_to_adata)s
        """
        from cellrank._utils._key import Key

        if self.adata is None:
            raise ValueError("Underlying annotated data object is not set.")

        key = Key.uns.kernel(self.backward, key=key)
        # retain the embedding info
        self.adata.uns[f"{key}_params"] = {
            **self.adata.uns.get(f"{key}_params", {}),
            **{"params": self.params},
            **{"init": self._init_kwargs},
        }
        self.adata.obsp[key] = self.transition_matrix.copy() if copy else self.transition_matrix

    @property
    def transition_matrix(self) -> Union[np.ndarray, sp.csr_matrix]:
        """Row-normalized transition matrix."""
        if self._parent is None and self._transition_matrix is None:
            self.compute_transition_matrix()
        return self._transition_matrix

    @transition_matrix.setter
    def transition_matrix(self, matrix: Tmat_t) -> None:
        """Set the transition matrix.

        Parameters
        ----------
        matrix
            Transition matrix. The matrix is row-normalized if necessary.

        Returns
        -------
        Nothing, just updates the :attr:`transition_matrix` and optionally normalizes it.
        """
        # fmt: off
        if matrix.shape != self.shape:
            raise ValueError(
                f"Expected matrix to be of shape `{self.shape}`, found `{matrix.shape}`."
            )

        def should_norm(mat: Tmat_t) -> bool:
            return not np.all(np.isclose(np.asarray(mat.sum(1)).squeeze(), 1.0, rtol=1e-12))

        if sp.issparse(matrix) and not sp.isspmatrix_csr(matrix):
            matrix = sp.csr_matrix(matrix)
        matrix = matrix.astype(np.float64, copy=False)

        force_normalize = (self._parent is None or self._normalize) and should_norm(matrix)
        if force_normalize:
            if np.any((matrix.data if sp.issparse(matrix) else matrix) < 0):
                raise ValueError("Unable to normalize matrix with negative values.")
            matrix = _normalize(matrix)
            if should_norm(matrix):  # some rows are all 0s/contain invalid values
                n_inv = np.sum(~np.isclose(np.asarray(matrix.sum(1)).squeeze(), 1.0, rtol=1e-12))
                raise ValueError(f"Transition matrix is not row stochastic, {n_inv} rows do not sum to 1.")
        # fmt: on

        self._transition_matrix = matrix

    @property
    def params(self) -> Dict[str, Any]:
        """Parameters which are used to compute the transition matrix."""
        if len(self.kernels) == 1:
            return self._params
        return {f"{k!r}:{i}": k.params for i, k in enumerate(self.kernels)}

    def _reuse_cache(self, expected_params: Dict[str, Any], *, time: Optional[Any] = None) -> bool:
        # fmt: off
        try:
            if expected_params == self._params:
                assert self.transition_matrix is not None
                logg.debug("Using cached transition matrix")
                logg.info("    Finish", time=time)
                return True
            return False
        except AssertionError:
            logg.warning("Transition matrix does not exist for the given parameters. Recomputing")
            return False
        except Exception as e:  # noqa: BLE001
            # e.g. the dict is not comparable
            logg.warning(f"Expected and actually parameters are not comparable, reason `{e}`. Recomputing")
            expected_params = {}  # clear the params
            return False
        finally:
            self._params = expected_params
        # fmt: on

    def _get_boundary(self, source: str, target: str, cluster_key: str, graph_key: str = "distances") -> List[int]:
        """Identify source observations at boundary to target cluster.

        Parameters
        ----------
        source
            Name of source cluster.
        target
            Name of target cluster.
        cluster_key
            Key in :attr:`~anndata.AnnData.obs` to obtain cluster annotations.
        graph_key
            Name of graph representation to use from :attr:`~anndata.AnnData.obsp`.

        Returns
        -------
        List of observation IDs at boundary to target cluster.
        """
        source_obs_mask = self.adata.obs[cluster_key].isin([source] if isinstance(source, str) else source)
        target_obs_mask = self.adata.obs[cluster_key].isin([target] if isinstance(target, str) else target)

        source_ids = np.where(source_obs_mask)[0]
        boundary_ids = []

        graph = self.adata.obsp[graph_key]
        for source_id in source_ids:
            obs_mask = graph[source_id, :].toarray().squeeze().astype(bool)

            if (obs_mask & target_obs_mask).any():
                boundary_ids.append(source_id)

        return boundary_ids

    def _get_empirical_velocity_field(
        self, boundary_ids: List[int], target_obs_mask, rep: str, graph_key: str = "distances"
    ) -> np.ndarray:
        """Compute an emprical estimate of velocity field between two clusters.

        Parameters
        ----------
        boundary_ids
            List of observation IDs at boundary to target cluster.
        target_obs_mask
            Boolean indicator identifying relevant observations from target.
        graph_key
            Name of graph representation to use from :attr:`~anndata.AnnData.obsp`.

        Returns
        -------
        Empirical velocity estimate.
        """
        obs_ids = np.arange(0, self.adata.n_obs)
        graph = self.adata.obsp[graph_key]
        features = self.adata.obsm[rep]
        empirical_velo = np.empty(shape=(len(boundary_ids), features.shape[1]))

        for idx, boundary_id in enumerate(boundary_ids):
            row = graph[boundary_id, :].toarray().squeeze()
            obs_mask = row.astype(bool) & target_obs_mask
            neighbors = obs_ids[obs_mask]
            weights = row[obs_mask]

            empirical_velo[idx, :] = np.sum(
                weights.reshape(-1, 1) * (features[neighbors, :] - features[boundary_id, :]), axis=0
            )

        empirical_velo = np.array(empirical_velo)
        obs_mask = np.isnan(empirical_velo).any(axis=1)
        empirical_velo = empirical_velo[~obs_mask, :]

        return empirical_velo

    def _get_vector_field_estimate(self, rep: str) -> np.ndarray:
        """Compute estimate of vector field under one step of the transition matrix.

        Parameters
        ----------
        rep
            Key in :attr:`~anndata.AnnData.obsm` to use as data representation.

        Returns
        -------
        Vector field estimate based on kernel dynamics.
        """
        extrapolated_gex = self.transition_matrix @ self.adata.obsm[rep]
        return extrapolated_gex - self.adata.obsm[rep]

    # TODO: Add definition/reference to paper
    def cbc(self, source: str, target: str, cluster_key: str, rep: str, graph_key: str = "distances") -> np.ndarray:
        """Compute cross-boundary correctness score between source and target cluster.

        Parameters
        ----------
        source
            Name of the source cluster.
        target
            Name of the target cluster.
        cluster_key
            Key in :attr:`~anndata.AnnData.obs` to obtain cluster annotations.
        rep
            Key in :attr:`~anndata.AnnData.obsm` to use as data representation.
        graph_key
            Name of graph representation to use from :attr:`~anndata.AnnData.obsp`.

        Returns
        -------
        Cross-boundary correctness score for each observation.
        """

        def _pearsonr(x: np.ndarray, y: np.ndarray) -> np.ndarray:
            x_centered = x - np.mean(x, axis=1, keepdims=True)
            y_centered = y - np.mean(y, axis=1, keepdims=True)
            denom = np.linalg.norm(x_centered, axis=1) * np.linalg.norm(y_centered, axis=1)

            return np.sum(x_centered * y_centered, axis=1) / denom

        target_obs_mask = self.adata.obs[cluster_key].isin([target])
        boundary_ids = self._get_boundary(source=source, target=target, cluster_key=cluster_key, graph_key=graph_key)
        empirical_velo = self._get_empirical_velocity_field(
            boundary_ids=boundary_ids, target_obs_mask=target_obs_mask, rep=rep, graph_key=graph_key
        )
        estimated_velo = self._get_vector_field_estimate(rep=rep)[boundary_ids, :]

        return _pearsonr(x=estimated_velo, y=empirical_velo)


@d.dedent
class Kernel(KernelExpression, abc.ABC):
    """Base kernel class.

    Parameters
    ----------
    %(adata)s
    parent
        Parent kernel expression.
    kwargs
        Keyword arguments for the parent.
    """

    def __init__(self, adata: AnnData, parent: Optional[KernelExpression] = None, **kwargs: Any):
        super().__init__(parent=parent, **kwargs)
        self._adata = adata
        self._n_obs = adata.n_obs
        self._read_from_adata(**kwargs)

    def _read_from_adata(self, **kwargs: Any) -> None:
        pass

    @property
    def adata(self) -> AnnData:
        """Annotated data object."""
        return self._adata

    @adata.setter
    def adata(self, adata: Optional[AnnData]) -> None:
        if adata is None:
            self._adata = None
            return
        if not isinstance(adata, AnnData):
            raise TypeError(f"Expected `adata` to be of type `AnnData`, found `{type(adata).__name__}`.")
        shape = (adata.n_obs, adata.n_obs)
        if self.shape != shape:
            raise ValueError(
                f"Expected new `AnnData` object to have same shape as the previous `{self.shape}`, found `{shape}`."
            )
        self._adata = adata

    @classmethod
    @d.dedent
    def from_adata(
        cls,
        adata: AnnData,
        key: str,
        copy: bool = False,
    ) -> "Kernel":
        """Read the kernel saved using :meth:`write_to_adata`.

        Parameters
        ----------
        %(adata)s
        key
            Key in :attr:`~anndata.AnnData.obsp` where the transition matrix is stored.
            The parameters should be stored in :attr:`adata.uns['{key}_params'] <anndata.AnnData.uns>`.
        copy
            Whether to copy the transition matrix.

        Returns
        -------
        The kernel with explicitly initialized properties:

        - :attr:`transition_matrix` - the transition matrix.
        - :attr:`params` - parameters used for computation.
        """
        transition_matrix = _read_graph_data(adata, key=key)
        try:
            params = adata.uns[f"{key}_params"]["params"].copy()
            init_params = adata.uns[f"{key}_params"]["init"].copy()
        except KeyError as e:
            raise KeyError(f"Unable to kernel parameters, reason: `{e}`") from e

        if copy:
            transition_matrix = transition_matrix.copy()

        kernel = cls(adata, **init_params)
        kernel.transition_matrix = transition_matrix
        kernel._params = params

        return kernel

    def copy(self, *, deep: bool = False) -> "Kernel":
        """Return a copy of self.

        Parameters
        ----------
        deep
            Whether to use :func:`~copy.deepcopy`.

        Returns
        -------
        Copy of self.
        """
        with self._remove_adata:
            k = copy.deepcopy(self)
        k.adata = self.adata.copy() if deep else self.adata
        return k

    def _copy_ignore(self, *attrs: str) -> "Kernel":
        # prevent copying attributes that are not necessary, e.g. during inversion
        sentinel, attrs = object(), set(attrs)
        objects = [
            (attr, obj)
            for attr, obj in ((attr, getattr(self, attr, sentinel)) for attr in attrs)
            if obj is not sentinel
        ]
        try:
            for attr, _ in objects:
                setattr(self, attr, None)
            return self.copy(deep=False)
        finally:
            for attr, obj in objects:
                setattr(self, attr, obj)

    def _format_params(self) -> str:
        n, _ = self.shape
        params = ", ".join(f"{k}={round(v, 3) if isinstance(v, float) else v!r}" for k, v in self.params.items())
        return f"n={n}, {params}" if params else f"n={n}"

    @property
    def transition_matrix(self) -> Union[np.ndarray, sp.csr_matrix]:
        """Row-normalized transition matrix."""
        return self._transition_matrix

    @transition_matrix.setter
    def transition_matrix(self, matrix: Any) -> None:
        KernelExpression.transition_matrix.fset(self, matrix)

    @property
    def kernels(self) -> Tuple["KernelExpression", ...]:
        """Underlying base kernels."""
        return (self,)

    @property
    def shape(self) -> Tuple[int, int]:
        """``(n_cells, n_cells)``."""
        return self._n_obs, self._n_obs

    def __getitem__(self, ix: int) -> "Kernel":
        if ix != 0:
            raise IndexError(ix)
        return self

    def __len__(self) -> int:
        return 1

    def __repr__(self) -> str:
        params = self._format_params()
        prefix = "~" if self.backward and self._parent is None else ""
        return f"{prefix}{self.__class__.__name__}[{params}]"

    def __str__(self) -> str:
        prefix = "~" if self.backward and self._parent is None else ""
        return f"{prefix}{self.__class__.__name__}[n={self.shape[0]}]"


class UnidirectionalKernel(UnidirectionalMixin, Kernel, abc.ABC):
    pass


class BidirectionalKernel(BidirectionalMixin, Kernel, abc.ABC):
    pass


class Constant(UnidirectionalKernel):
    def __init__(self, adata: AnnData, value: Union[int, float]):
        super().__init__(adata)
        self.transition_matrix = value

    def compute_transition_matrix(self, value: Union[int, float]) -> "Constant":
        self.transition_matrix = value
        return self

    @property
    def transition_matrix(self) -> Union[int, float]:
        return self._transition_matrix

    @transition_matrix.setter
    def transition_matrix(self, value: Union[int, float]) -> None:
        if not isinstance(value, (int, float, np.integer, np.floating)):
            raise TypeError(f"Value must be a `float` or `int`, found `{type(value).__name__}`.")
        if value <= 0:
            raise ValueError(f"Expected the scalar to be positive, found `{value}`.")

        self._transition_matrix = value
        self._params = {"value": value}

    def copy(self, *, deep: bool = False) -> "Constant":
        return Constant(self.adata, self.transition_matrix)

    # fmt: off
    def __radd__(self, other: Union[int, float, "KernelExpression"]) -> "Constant":
        if isinstance(other, (int, float, np.integer, np.floating)):
            other = Constant(self.adata, other)
        if isinstance(other, Constant):
            if self.shape != other.shape:
                raise ValueError(f"Expected kernel shape to be `{self.shape}`, found `{other.shape}`.")
            return Constant(self.adata, self.transition_matrix + other.transition_matrix)

        return super().__radd__(other)

    def __rmul__(self, other: Union[int, float, "KernelExpression"]) -> "Constant":
        if isinstance(other, (int, float, np.integer, np.floating)):
            other = Constant(self.adata, other)
        if isinstance(other, Constant):
            if self.shape != other.shape:
                raise ValueError(f"Expected kernel shape to be `{self.shape}`, found `{other.shape}`.")
            return Constant(self.adata, self.transition_matrix * other.transition_matrix)

        return super().__rmul__(other)
    # fmt: on

    def __repr__(self) -> str:
        return repr(round(self.transition_matrix, 3))

    def __str__(self) -> str:
        return str(round(self.transition_matrix, 3))


class NaryKernelExpression(BidirectionalMixin, KernelExpression):
    def __init__(self, *kexprs: KernelExpression, parent: Optional[KernelExpression] = None):
        super().__init__(parent=parent)
        self._validate(kexprs)

    @abc.abstractmethod
    def _combine_transition_matrices(self, t1: Tmat_t, t2: Tmat_t) -> Tmat_t:
        pass

    @property
    @abc.abstractmethod
    def _initial_value(self) -> Union[int, float, Tmat_t]:
        pass

    @property
    @abc.abstractmethod
    def _combiner(self) -> str:
        pass

    def compute_transition_matrix(self) -> "NaryKernelExpression":
        for kexpr in self:
            if kexpr.transition_matrix is None:
                if isinstance(kexpr, Kernel):
                    raise RuntimeError(
                        f"`{kexpr}` is uninitialized. Compute its "
                        f"transition matrix first as `.compute_transition_matrix()`."
                    )
                kexpr.compute_transition_matrix()

        tmat = self._initial_value
        for kexpr in self:
            tmat = self._combine_transition_matrices(tmat, kexpr.transition_matrix)
        self.transition_matrix = tmat

        return self

    def _validate(self, kexprs: Sequence[KernelExpression]) -> None:
        if not len(kexprs):
            raise ValueError("No kernels to combined.")

        shapes = {kexpr.shape for kexpr in kexprs}
        if len(shapes) > 1:
            raise ValueError(f"Expected all kernels to have the same shapes, found `{sorted(shapes)}`.")

        directions = {kexpr.backward for kexpr in kexprs}
        if True in directions and False in directions:
            raise ValueError("Unable to combine both forward and backward kernels.")

        self._backward = True if True in directions else False if False in directions else None
        self._kexprs = kexprs
        for kexpr in self:
            kexpr._parent = self

    @property
    def adata(self) -> AnnData:
        """Annotated data object."""
        return self[0].adata

    @adata.setter
    def adata(self, adata: Optional[AnnData]) -> None:
        # allow resetting (use for temp. pickling without adata)
        for kexpr in self:
            kexpr.adata = adata

    @property
    def kernels(self) -> Tuple["KernelExpression", ...]:
        """Underlying unique basic kernels."""
        kernels = []
        for kexpr in self:
            if isinstance(kexpr, Kernel) and not isinstance(kexpr, Constant):
                kernels.append(kexpr)
            elif isinstance(kexpr, NaryKernelExpression):  # recurse
                kernels.extend(kexpr.kernels)

        # return only unique kernels
        return tuple(set(kernels))

    def copy(self, *, deep: bool = False) -> "KernelExpression":
        kexprs = (k.copy(deep=deep) for k in self)
        return type(self)(*kexprs, parent=self._parent)

    @property
    def shape(self) -> Tuple[int, int]:
        """``(n_cells, n_cells)``."""
        # all kernels have the same shape
        return self[0].shape

    def _format(self, formatter: Callable[[KernelExpression], str]) -> str:
        return (
            f"{'~' if self.backward and self._parent is None else ''}("
            + f" {self._combiner} ".join(formatter(kexpr) for kexpr in self)
            + ")"
        )

    def __getitem__(self, ix: int) -> "KernelExpression":
        return self._kexprs[ix]

    def __len__(self) -> int:
        return len(self._kexprs)

    def __invert__(self) -> "KernelExpression":
        if self.backward is None:
            return self.copy()
        kexprs = tuple(~k if isinstance(k, BidirectionalMixin) else k.copy() for k in self)
        kexpr = type(self)(*kexprs, parent=self._parent)
        kexpr._transition_matrix = None
        return kexpr

    def __repr__(self) -> str:
        return self._format(repr)

    def __str__(self) -> str:
        return self._format(str)


class KernelAdd(NaryKernelExpression):
    def compute_transition_matrix(self) -> "KernelAdd":
        self._maybe_recalculate_constants()
        return super().compute_transition_matrix()

    def _combine_transition_matrices(self, t1: Tmat_t, t2: Tmat_t) -> Tmat_t:
        return t1 + t2

    def _maybe_recalculate_constants(self) -> None:
        """Normalize constants to sum to 1."""
        constants = self._bin_consts
        if constants:
            total = sum((c.transition_matrix for c in constants), 0.0)
            for c in constants:
                c.transition_matrix = c.transition_matrix / total
            for kexpr in self:  # don't normalize  (c * x)
                kexpr._normalize = False

    @property
    def _bin_consts(self) -> List[KernelExpression]:
        """Return constant expressions for each binary multiplication children with at least 1 constant."""
        return [c for k in self if isinstance(k, KernelMul) for c in k._bin_consts]

    @property
    def _combiner(self) -> str:
        return "+"

    @property
    def _initial_value(self) -> Union[int, float, Tmat_t]:
        return 0.0


class KernelMul(NaryKernelExpression):
    def _combine_transition_matrices(self, t1: Tmat_t, t2: Tmat_t) -> Tmat_t:
        if sp.issparse(t1):
            return t1.multiply(t2)
        if sp.issparse(t2):
            return t2.multiply(t1)
        return t1 * t2

    def _format(self, formatter: Callable[[KernelExpression], str]) -> str:
        fmt = super()._format(formatter)
        if fmt[0] == "~" or not self._bin_consts:
            return fmt
        if fmt.startswith("("):
            fmt = fmt[1:]
        if fmt.endswith(")"):
            fmt = fmt[:-1]
        return fmt

    @property
    def _bin_consts(self) -> List[KernelExpression]:
        """Return all constants if this expression contains only 2 subexpressions."""
        if len(self) != 2:
            return []

        return [k for k in self if isinstance(k, Constant)]

    @property
    def _split_const(self) -> Tuple[Optional[Constant], Optional[KernelExpression]]:
        """Return a constant and the other expression, iff this expression is of length 2 and contains a constant."""
        if not self._bin_consts:
            return None, None
        k1, k2 = self
        if isinstance(k1, Constant):
            return k1, k2
        return k2, k1

    @property
    def _combiner(self) -> str:
        return "*"

    @property
    def _initial_value(self) -> Union[int, float, Tmat_t]:
        return 1.0



import collections
import copy
import itertools
import pathlib
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

import numpy as np
import pandas as pd
from pandas.api.types import infer_dtype

import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib import cm, colors
from mpl_toolkits.axes_grid1 import make_axes_locatable

from anndata import AnnData

from cellrank import logging as logg
from cellrank._utils._colors import _create_categorical_colors
from cellrank._utils._docs import d
from cellrank._utils._enum import DEFAULT_BACKEND
from cellrank._utils._parallelize import parallelize
from cellrank._utils._utils import _unique_order_preserving, save_fig
from cellrank.models import GAMR, BaseModel, FailedModel, SKLearnModel
from cellrank.models._base_model import ColorType

__all__ = ["composition"]

Queue = TypeVar("Queue")
Graph = TypeVar("Graph")


_ERROR_INCOMPLETE_SPEC = "No options were specified for {}. " "Consider specifying a fallback model using '*'."
_time_range_type = Optional[Union[float, Tuple[Optional[float], Optional[float]]]]
_return_model_type = Mapping[str, Mapping[str, BaseModel]]
_input_model_type = Union[BaseModel, _return_model_type]
_callback_type = Optional[Union[Callable, Mapping[str, Mapping[str, Callable]]]]


class BulkRes(NamedTuple):
    x_test: np.ndarray
    y_test: np.ndarray


def _is_any_gam_mgcv(models: Union[BaseModel, Dict[str, Dict[str, BaseModel]]]) -> bool:
    """Return whether any models to be fit are from R's `mgcv` package.

    Parameters
    ----------
    models
        Model used for fitting.

    Returns
    -------
    :obj:`True` if any of the models is from R's mgcv package, else :obj:`False`.
    """
    return isinstance(models, GAMR) or (
        isinstance(models, dict) and any(isinstance(m, GAMR) for ms in models.values() for m in ms.values())
    )


def _create_models(
    model: _input_model_type, obs: Sequence[str], lineages: Sequence[Optional[str]]
) -> _return_model_type:
    """Create models for each gene and lineage.

    Parameters
    ----------
    obs
        Sequence of observations, such as genes.
    lineages
        Sequence of genes.

    Returns
    -------
    The created models.
    """

    def process_lineages(obs_name: str, lin_names: Union[BaseModel, Dict[Optional[str], Any]]):
        if isinstance(lin_names, BaseModel):
            # sharing the same models for all lineages
            for lin_name in lineages:
                models[obs_name][lin_name] = copy.copy(lin_names)
            return
        if not isinstance(lin_names, dict):
            raise TypeError(
                f"Expected the model to be either a lineage specific `dict` or a `BaseModel`, "
                f"found `{type(lin_names).__name__!r}`."
            )

        lin_rest_model = lin_names.get("*", None)  # do not pop
        if lin_rest_model is not None and not isinstance(lin_rest_model, BaseModel):
            raise TypeError(
                f"Expected the lineage fallback model for gene `{obs_name!r}` to be of type `BaseModel`, "
                f"found `{type(lin_rest_model).__name__!r}`."
            )

        for lin_name, mod in lin_names.items():
            if lin_name == "*":
                continue
            if not isinstance(mod, BaseModel):
                raise TypeError(
                    f"Expected the model for gene `{obs_name!r}` and lineage `{lin_name!r}` "
                    f"to be of type `BaseModel`, found `{type(mod).__name__!r}`."
                )
            models[obs_name][lin_name] = copy.copy(mod)

        if set(models[obs_name].keys()) & lineages == lineages:
            return

        if lin_rest_model is not None:
            for lin_name in lineages - set(models[obs_name].keys()):
                models[obs_name][lin_name] = copy.copy(lin_rest_model)
        else:
            raise ValueError(_ERROR_INCOMPLETE_SPEC.format(f"all lineages for gene `{obs_name!r}`"))

    if not len(lineages):
        raise ValueError("No lineages have been selected.")

    if not len(obs):
        raise ValueError("No genes have been selected.")

    if isinstance(model, BaseModel):
        return {
            o: {lin: copy.copy(model) for lin in _unique_order_preserving(lineages)}
            for o in _unique_order_preserving(obs)
        }

    lineages, obs = (
        set(_unique_order_preserving(lineages)),
        set(_unique_order_preserving(obs)),
    )
    models = collections.defaultdict(dict)

    if isinstance(model, dict):
        obs_rest_model = model.pop("*", None)
        if obs_rest_model is not None and not isinstance(obs_rest_model, BaseModel):
            raise TypeError(
                f"Expected the gene fallback model to be of type `BaseModel`, "
                f"found `{type(obs_rest_model).__name__!r}`."
            )

        for obs_name, lin_names in model.items():
            process_lineages(obs_name, lin_names)

        if obs_rest_model is not None:
            for obs_name in obs - set(model.keys()):
                process_lineages(obs_name, model.get(obs_name, obs_rest_model))
        elif set(model.keys()) != obs:
            raise ValueError(_ERROR_INCOMPLETE_SPEC.format(f"genes `{list(obs - set(model.keys()))}`."))
    else:
        raise TypeError(
            f"Class `{type(model).__name__!r}` must be of type `BaseModel` or "
            f"a gene and lineage specific `dict` of `BaseModel`.."
        )

    if set(models.keys()) & obs != obs:
        raise ValueError(f"Missing gene models for the following genes: `{list(obs - set(models.keys()))}`.")

    for gene, vs in models.items():
        if set(vs.keys()) & lineages != lineages:
            raise ValueError(f"Missing lineage models for the gene `{gene!r}`: `{list(lineages - set(vs.keys()))}`.")

    return models


def _fit_bulk_helper(
    genes: Sequence[str],
    models: _input_model_type,
    callbacks: _callback_type,
    lineages: Sequence[Optional[str]],
    time_range: Sequence[Union[float, Tuple[float, float]]],
    return_models: bool = False,
    queue: Optional[Queue] = None,
    **kwargs,
) -> Dict[str, Dict[str, BaseModel]]:
    """Fit model for given genes and lineages.

    Parameters
    ----------
    genes
        Genes for which to fit the models.
    models
        Gene and lineage specific models.
    callbacks
        Gene and lineage specific prepare callbacks.
    lineages
        Lineages for which to fit the models.
    time_range
        Minimum and maximum pseudotimes.
    return_models
        Whether to return the full models or just tuple ``(x_test,  y_test)``.
    queue
        Signalling queue in the parent process/thread used to update the progress bar.
    kwargs
        Keyword arguments for :func:`~cellrank.models.BaseModel.prepare`.

    Returns
    -------
    The fitted models, optionally containing the confidence interval in the form of
    ``{'gene1': {'lineage1': <model11>, ...}, ...}``.
    If any step has failed, the model will be of type :class:`~cellrank.models.FailedModel`.
    """
    if len(lineages) != len(time_range):
        raise ValueError(
            f"Expected `lineage` and `time_range` to be of same length, "
            f"found `{len(lineages)}` != `{len(time_range)}`."
        )

    conf_int = return_models and kwargs.pop("conf_int", False)
    res = {}

    for gene in genes:
        res[gene] = {}
        for ln, tr in zip(lineages, time_range):
            cb = callbacks[gene][ln]
            model = models[gene][ln]
            model._is_bulk = True

            model = cb(model, gene=gene, lineage=ln, time_range=tr, **kwargs)
            model = model.fit()
            # GAMR is a bit faster if we don't need the conf int
            # if it's needed, `.predict` will calculate it and `confidence_interval` will do nothing
            if not conf_int:
                model.predict()
            elif _is_any_gam_mgcv(model):
                model.predict(level=conf_int if isinstance(conf_int, float) else 0.95)
            else:
                model.predict()
                model.confidence_interval()

            res[gene][ln] = model if return_models else BulkRes(model.x_test, model.y_test)

        if queue is not None:
            queue.put(1)

    if queue is not None:
        queue.put(None)

    return res


def _fit_bulk(
    models: Mapping[str, Mapping[str, Callable]],
    callbacks: Mapping[str, Mapping[str, Callable]],
    genes: Union[str, Sequence[str]],
    lineages: Union[str, Sequence[str]],
    time_range: _time_range_type,
    parallel_kwargs: dict,
    return_models: bool = False,
    filter_all_failed: bool = True,
    **kwargs,
) -> Tuple[_return_model_type, _return_model_type, Sequence[str], Sequence[str]]:
    """Fit models for given genes and lineages.

    Parameters
    ----------
    models
        Gene and lineage specific estimators.
    callbacks
        Functions which are called to prepare the ``models``.
    genes
        Genes for which to fit the ``models``.
    lineages
        Lineages for which to fit the ``models``.
    time_range
        Possibly ``lineages`` specific start- and endtimes.
    parallel_kwargs
        Keyword arguments for :func:`~cellrank._utils._parallelize.parallelize`.
    return_models
        Whether to return the full models or just a dictionary of dictionaries of :class:`~typing.NamedTuple`,
        ``(x_test, y_test)``. This is highly discouraged because no meaningful error messages will be produced.
    filter_all_failed
        Whether to filter out all models which have failed.

    Returns
    -------
    All the models, including the failed ones. It is a nested dictionary where keys are the ``genes`` and the values
    is again a :class:`dict`, where keys are ``lineages`` and values are the failed or fitted models or
    the :class:`collections.namedtuple`, based on ``return_models = True``.

    Same as above, but can contain failed models if ``filter_all_failed=False``. In that case, it is guaranteed
    that this dictionary will contain only genes which have been successfully fitted for at least 1 lineage.
    If ``return_models = True``, the models are just a :class:`collections.namedtuple` of `(x_test, y_test)`.

    All the genes of the filtered models.

    All the lineage of the filtered models.
    """
    if isinstance(genes, str):
        genes = [genes]
    if isinstance(lineages, str):
        lineages = [lineages]

    if isinstance(time_range, (tuple, float, int, type(None))):
        time_range = [time_range] * len(lineages)
    elif len(time_range) != len(lineages):
        raise ValueError(f"Expected time ranges to be of length `{len(lineages)}`, found `{len(time_range)}`.")

    n_jobs = parallel_kwargs.pop("n_jobs", 1)

    start = logg.info(f"Computing trends using `{n_jobs}` core(s)")
    models = parallelize(
        _fit_bulk_helper,
        genes,
        unit="gene" if kwargs.get("data_key", "gene") != "obs" else "obs",
        n_jobs=n_jobs,
        extractor=lambda modelss: {k: v for m in modelss for k, v in m.items()},
    )(
        models=models,
        callbacks=callbacks,
        lineages=lineages,
        time_range=time_range,
        return_models=return_models,
        **kwargs,
    )
    logg.info("    Finish", time=start)

    return _filter_models(models, return_models=return_models, filter_all_failed=filter_all_failed)


def _filter_models(
    models, return_models: bool = False, filter_all_failed: bool = True
) -> Tuple[_return_model_type, _return_model_type, Sequence[str], Sequence[str]]:
    def is_valid(x: Union[BaseModel, BulkRes]) -> bool:
        if return_models:
            assert isinstance(x, BaseModel), f"Expected `BaseModel`, found `{type(x).__name__!r}`."
            return bool(x)

        return x.x_test is not None and x.y_test is not None and np.all(np.isfinite(x.y_test))

    modelmat = pd.DataFrame(models).T
    # `DataFrame.map` in `pandas>=2.1.0`
    modelmask = modelmat.map(is_valid) if hasattr(modelmat, "map") else modelmat.applymap(is_valid)
    to_keep = modelmask[modelmask.any(axis=1)]
    to_keep = to_keep.loc[:, to_keep.any(axis=0)].T

    filtered_models = {
        gene: {
            ln: models[gene][ln]
            for ln in (ln for ln in v if (is_valid(models[gene][ln]) if filter_all_failed else True))
        }
        for gene, v in to_keep.to_dict().items()
    }

    if not len(filtered_models):
        if not return_models:
            raise RuntimeError(
                "Fitting has failed for all gene/lineage combinations. "
                "Specify `return_models=True` for more information."
            )
        for ms in models.values():
            for model in ms.values():
                assert isinstance(model, FailedModel), f"Expected `FailedModel`, found `{type(model).__name__!r}`."
                model.reraise()

    if not np.all(modelmask.values):
        failed_models = modelmat.values[~modelmask.values]
        logg.warning(
            f"Unable to fit `{len(failed_models)}` models." + ""
            if return_models
            else "Consider specify `return_models=True` for further inspection."
        )
        logg.debug("The failed models were:\n`{}`".format("\n".join(f"    {m}" for m in failed_models)))

    # lineages is the max number of lineages
    return models, filtered_models, tuple(filtered_models.keys()), tuple(to_keep.index)


@d.dedent
def _trends_helper(
    models: Dict[str, Dict[str, Any]],
    gene: str,
    transpose: bool = False,
    lineage_names: Optional[Sequence[str]] = None,
    same_plot: bool = False,
    sharey: Union[str, bool] = False,
    show_ylabel: bool = True,
    show_lineage: Union[bool, np.ndarray] = True,
    show_xticks_and_label: Union[bool, np.ndarray] = True,
    lineage_cmap: Optional[Union[mpl.colors.ListedColormap, Sequence]] = None,
    lineage_probability_color: Optional[str] = None,
    fate_prob_cmap=cm.viridis,
    gene_as_title: bool = False,
    cell_color: Optional[str] = None,
    legend_loc: Optional[str] = "best",
    fig: mpl.figure.Figure = None,
    axes: Union[mpl.axes.Axes, Sequence[mpl.axes.Axes]] = None,
    **kwargs: Any,
) -> None:
    """Plot an expression gene for some lineages.

    Parameters
    ----------
    %(adata)s
    %(model)s
    gene
        Name of the gene in :attr:`~anndata.AnnData.var_names`.
    ln_key
        Key in :attr:`~anndata.AnnData.bosm` where to find the lineages.
    lineage_names
        Names of lineages to plot.
    same_plot
        Whether to plot all lineages in the same plot or separately.
    sharey
        Whether the y-axis is being shared.
    show_ylabel
        Whether to show y-label on the y-axis. Usually, only the first column will contain the y-label.
    show_lineage
        Whether to show the lineage as the title. Usually, only first row will contain the lineage names.
    show_xticks_and_label
        Whether to show x-ticks and x-label. Usually, only the last row will show this.
    lineage_cmap
        Colormap to use when coloring the lineage. When ``transpose = True``, this corresponds to the color of genes.
    lineage_probability_color
        Actual color of one ``lineage``. Only used when ``same_plot = True`` and ``transpose = True`` and
        ``lineage_probability = True``.
    fate_prob_cmap:
        Colormap to use when coloring in the fate probabilities, if they are being plotted.
    gene_as_title
        Whether to use the gene names as titles (with lineage names as well) or on the y-axis.
    legend_loc
        Location of the legend. If :obj:`None`, don't show any legend.
    fig
        Figure to use.
    ax
        Ax to use.
    kwargs
        Keyword arguments for :meth:`~cellrank.models.BaseModel.plot`.

    Returns
    -------
    %(just_plots)s
    """
    n_lineages = len(lineage_names)
    if same_plot:
        axes = [axes] * len(lineage_names)

    axes = np.ravel(axes)

    percs = kwargs.pop("perc", None)
    if percs is None or not isinstance(percs[0], (tuple, list)):
        percs = [percs]

    same_perc = False  # we need to show colorbar always if percs differ
    if len(percs) != n_lineages or n_lineages == 1:
        if len(percs) != 1:
            raise ValueError(f"Percentile must be a collection of size `1` or `{n_lineages}`, got `{len(percs)}`.")
        same_perc = True
        percs = percs * n_lineages

    hide_cells = kwargs.pop("hide_cells", False)
    show_cbar = kwargs.pop("cbar", True)
    show_prob = kwargs.pop("lineage_probability", False)

    if same_plot:
        if not transpose:
            lineage_colors = (
                lineage_cmap.colors if lineage_cmap is not None and hasattr(lineage_cmap, "colors") else lineage_cmap
            )
        else:
            # this should be fine w.r.t. to the missing genes, since they are in the same order AND
            # we're also passing the failed models (this is important)
            # these are actually gene colors
            if lineage_cmap is not None:
                lineage_colors = (
                    lineage_cmap.colors
                    if hasattr(lineage_cmap, "colors")
                    else [c for _, c in zip(lineage_names, lineage_cmap)]
                )
            else:
                lineage_colors = _create_categorical_colors(n_lineages)
    else:
        lineage_colors = (("black" if not colors.is_color_like(lineage_cmap) else lineage_cmap),) * n_lineages

    if n_lineages > len(lineage_colors):
        raise ValueError(f"Expected at least `{n_lineages}` colors, found `{len(lineage_colors)}`.")

    lineage_color_mapper = {ln: lineage_colors[i] for i, ln in enumerate(lineage_names)}
    successful_models = {ln: models[gene][ln] for ln in lineage_names if models[gene][ln]}

    if show_prob and same_plot:
        minns, maxxs = zip(
            *(
                models[gene][n]._return_min_max(
                    show_conf_int=kwargs.get("conf_int", False),
                )
                for n in lineage_names
            )
        )
        minn, maxx = min(minns), max(maxxs)
        kwargs["loc"] = legend_loc
        kwargs["scaler"] = lambda x: (x - minn) / (maxx - minn)
    else:
        kwargs["loc"] = None

    if isinstance(show_xticks_and_label, bool):
        show_xticks_and_label = [show_xticks_and_label] * len(lineage_names)
    elif len(show_xticks_and_label) != len(lineage_names):
        raise ValueError(
            f"Expected `show_xticks_label` to be the same length as `lineage_names`, "
            f"found `{len(show_xticks_and_label)}` != `{len(lineage_names)}`."
        )

    if isinstance(show_lineage, bool):
        show_lineage = [show_lineage] * len(lineage_names)
    elif len(show_lineage) != len(lineage_names):
        raise ValueError(
            f"Expected `show_lineage` to be the same length as `lineage_names`, "
            f"found `{len(show_lineage)}` != `{len(lineage_names)}`."
        )

    last_ax = None
    ylabel_shown = False
    cells_shown = False
    obs_legend_loc = kwargs.pop("obs_legend_loc", "best")

    for i, (name, ax, perc) in enumerate(zip(lineage_names, axes, percs)):
        model = models[gene][name]
        if isinstance(model, FailedModel):
            if not same_plot:
                ax.remove()
            continue

        if same_plot:
            if gene_as_title:
                title = gene
                ylabel = "expression" if show_ylabel else None
            else:
                title = ""
                ylabel = gene
        else:
            if gene_as_title:
                title = None
                ylabel = "expression" if not ylabel_shown else None
            else:
                title = (name if name is not None else "no lineage") if show_lineage[i] else ""
                ylabel = gene if not ylabel_shown else None

        model.plot(
            ax=ax,
            fig=fig,
            perc=perc,
            cell_color=cell_color,
            cbar=False,
            obs_legend_loc=None,
            title=title,
            hide_cells=True if hide_cells else cells_shown if same_plot else False,
            same_plot=same_plot,
            lineage_color=lineage_color_mapper[name],
            lineage_probability_color=lineage_probability_color,
            fate_prob_cmap=fate_prob_cmap,
            lineage_probability=show_prob,
            ylabel=ylabel,
            **kwargs,
        )
        if sharey in ("row", "all", True) and not ylabel_shown:
            plt.setp(ax.get_yticklabels(), visible=True)

        if show_xticks_and_label[i]:
            plt.setp(ax.get_xticklabels(), visible=True)
        else:
            ax.set_xlabel(None)

        last_ax = ax
        ylabel_shown = True
        cells_shown = True

    key, color, typp, mapper = model._get_colors(cell_color, same_plot=same_plot)
    if typp == ColorType.CAT:
        if not hide_cells:
            model._maybe_add_legend(fig, ax, mapper=mapper, title=key, loc=obs_legend_loc, is_line=False)
    elif typp == ColorType.CONT and same_perc and show_cbar and not hide_cells:
        if isinstance(color, np.ndarray):
            # plotting cont. observation other than lin. probs as a color
            vmin = np.min(color)
            vmax = np.max(color)
        else:
            vmin = np.min([model.w_all for model in successful_models.values()])
            vmax = np.max([model.w_all for model in successful_models.values()])
        norm = colors.Normalize(vmin=vmin, vmax=vmax)

        for ax in axes:
            children = [c for c in ax.get_children() if isinstance(c, mpl.collections.PathCollection)]
            if len(children):
                children[0].set_norm(norm)

        divider = make_axes_locatable(last_ax)
        cax = divider.append_axes("right", size="2%", pad=0.1)
        _ = mpl.colorbar.ColorbarBase(
            cax,
            norm=norm,
            cmap=fate_prob_cmap,
            label=key,
            ticks=np.linspace(norm.vmin, norm.vmax, 5),
        )

    if same_plot and lineage_names != [None]:
        model._maybe_add_legend(
            fig,
            ax,
            mapper={ln: lineage_color_mapper[ln] for ln in successful_models},
            loc=legend_loc,
        )


def _position_legend(ax: mpl.axes.Axes, legend_loc: str, **kwargs) -> mpl.legend.Legend:
    """Position legend inside or outside the figure.

    Parameters
    ----------
    ax
        Ax where to position the legend.
    legend_loc
        Position of legend.
    kwargs
        Keyword arguments for :meth:`matplotlib.axes.Axes.legend`.

    Returns
    -------
    The created legend.
    """
    if legend_loc == "center center out":
        raise ValueError("Invalid option: `'center center out'`.")
    if legend_loc == "best":
        return ax.legend(loc="best", **kwargs)

    tmp, loc = legend_loc.split(" "), ""

    if len(tmp) == 1:
        height, rest = tmp[0], []
        width = "right" if height in ("upper", "top", "center") else "left"
    else:
        height, width, *rest = legend_loc.split(" ")
        if rest:
            if len(rest) != 1:
                raise ValueError(f"Expected only 1 additional modifier ('in' or 'out'), found `{list(rest)}`.")
            if rest[0] not in ("in", "out"):
                raise ValueError(f"Invalid modifier `{rest[0]!r}`. Valid options are: `'in', 'out'`.")
            if rest[0] == "in":  # ignore in, it's default
                rest = []

    if height in ("upper", "top"):
        y = 1.55 if width == "center" else 1.025
        loc += "upper"
    elif height == "center":
        y = 0.5
        loc += "center"
    elif height in ("lower", "bottom"):
        y = -0.55 if width == "center" else -0.025
        loc += "lower"
    else:
        raise ValueError(
            f"Invalid legend position on y-axis: `{height!r}`. "
            f"Valid options are: `'upper', 'top', 'center', 'lower', 'bottom'`."
        )

    if width == "left":
        x = -0.05
        loc += " right" if rest else " left"
    elif width == "center":
        x = 0.5
        if height != "center":  # causes to be like top center
            loc += " center"
    elif width == "right":
        x = 1.05
        loc += " left" if rest else " right"
    else:
        raise ValueError(
            f"Invalid legend position on x-axis: `{width!r}`. " f"Valid options are: `'left', 'center', 'right'`."
        )

    if rest:
        kwargs["bbox_to_anchor"] = (x, y)

    return ax.legend(loc=loc, **kwargs)


def _get_backend(model, backend: str) -> str:
    return DEFAULT_BACKEND if _is_any_gam_mgcv(model) else backend


@d.dedent
def _create_callbacks(
    adata: AnnData,
    callback: Optional[Callable],
    obs: Sequence[str],
    lineages: Sequence[Optional[str]],
    perform_sanity_check: Optional[bool] = None,
    **kwargs,
) -> Dict[str, Dict[str, Callable]]:
    """Create models for each gene and lineage.

    Parameters
    ----------
    %(adata)s
    callback
        Gene and lineage specific prepare callbacks.
    obs
        Sequence of observations, such as genes.
    lineages
        Sequence of genes.
    perform_sanity_check
        Whether to check if all callbacks have the correct signature. This is done by instantiating
        dummy model and running the function. We're assuming that the callback isn't really a pricey operation.
        If :obj:`None`, it is only performed for non-default callbacks.
    kwargs
        Keyword arguments for ``callback`` when performing the sanity check.

    Returns
    -------
    The created callbacks.
    """

    def process_lineages(obs_name: str, lin_names: Optional[Union[Callable, Dict[Optional[str], Any]]]) -> None:
        if lin_names is None:
            lin_names = _default_model_callback

        if callable(lin_names):
            # sharing the same models for all lineages
            for lin_name in lineages:
                callbacks[obs_name][lin_name] = lin_names
            return
        if not isinstance(lin_names, dict):
            raise TypeError(
                f"Expected the lineage callback to be either `callable` or a dictionary of callables, "
                f"found `{type(lin_names).__name__!r}`."
            )

        lin_rest_callback = lin_names.get("*", _default_model_callback) or _default_model_callback  # do not pop
        if not callable(lin_rest_callback):
            raise TypeError(
                f"Expected the lineage fallback callback for gene `{obs_name!r}` to be `callable`, "
                f"found `{type(lin_rest_callback).__name__!r}`."
            )

        for lin_name, cb in lin_names.items():
            if lin_name == "*":
                continue
            if not callable(cb):
                raise TypeError(
                    f"Expected the callback for gene `{obs_name!r}` and lineage `{lin_name!r}` "
                    f"to be `callable`, found `{type(cb).__name__!r}`."
                )
            callbacks[obs_name][lin_name] = cb

        for lin_name in lineages - set(callbacks[obs_name].keys()):
            callbacks[obs_name][lin_name] = lin_rest_callback

    def maybe_sanity_check(callbacks: Dict[str, Dict[str, Callable]]) -> None:
        if not perform_sanity_check:
            return

        from sklearn.svm import SVR

        logg.debug("Performing callback sanity checks")
        for gene in callbacks:
            for lineage, cb in callbacks[gene].items():
                # create the model here because the callback can search the attribute
                dummy_model = SKLearnModel(adata, model=SVR())
                try:
                    model = cb(dummy_model, gene=gene, lineage=lineage, **kwargs)
                    assert model is dummy_model, (
                        "Creation of new models is not allowed. " "Ensure that callback returns the same model."
                    )
                    assert model.prepared, "Model is not prepared. Ensure that callback calls `.prepare()`."
                    assert model._gene == gene, f"Callback modified the gene from `{gene!r}` to `{model._gene!r}`."
                    assert (
                        model._lineage == lineage
                    ), f"Callback modified the lineage from `{lineage!r}` to `{model._lineage!r}`."
                    if isinstance(model, FailedModel):
                        model.reraise()
                except Exception as e:  # noqa: BLE001
                    raise RuntimeError(
                        f"Callback validation failed for gene `{gene!r}` and lineage `{lineage!r}`."
                    ) from e

    def all_callbacks_are_default(cbs: dict) -> bool:
        # this correctly implicitly handles '*': None
        for vs in cbs.values():
            if isinstance(vs, dict):
                for cb in vs.values():
                    if callable(cb) and cb is not _default_model_callback:
                        return False
            elif callable(vs) and vs is not _default_model_callback:
                return False

        return True

    if not len(lineages):
        raise ValueError("No lineages have been selected.")

    if not len(obs):
        raise ValueError("No genes have been selected.")

    if callback is None:
        callback = _default_model_callback

    if perform_sanity_check is None:
        perform_sanity_check = (
            not all_callbacks_are_default(callback)
            if isinstance(callback, dict)
            else callback is not _default_model_callback
        )

    if callable(callback):
        callbacks = {o: {lin: callback for lin in lineages} for o in obs}
        maybe_sanity_check(callbacks)
        return callbacks

    lineages, obs = (
        set(_unique_order_preserving(lineages)),
        set(_unique_order_preserving(obs)),
    )
    callbacks = collections.defaultdict(dict)

    if isinstance(callback, dict):
        # can be specified as None
        obs_rest_callback = callback.pop("*", _default_model_callback) or _default_model_callback

        for obs_name, lin_names in callback.items():
            process_lineages(obs_name, lin_names)

        if callable(obs_rest_callback):
            for obs_name in obs - set(callback.keys()):
                process_lineages(obs_name, callback.get(obs_name, obs_rest_callback))
        else:
            raise TypeError(
                f"Expected the gene fallback callback to be `callable`, "
                f"found `{type(obs_rest_callback).__name__!r}`."
            )
    else:
        raise TypeError(
            f"Class `{type(callback).__name__!r}` must be `callable` or "
            f"a gene and lineage specific `dict` of `callables`."
        )

    if set(callbacks.keys()) & obs != obs:
        raise ValueError(f"Missing gene callbacks for the following genes: `{list(obs - set(callbacks.keys()))}`.")

    for gene, vs in callbacks.items():
        if set(vs.keys()) & lineages != lineages:
            raise ValueError(f"Missing lineage callbacks for gene `{gene!r}`: `{list(lineages - set(vs.keys()))}`.")

    maybe_sanity_check(callbacks)

    return callbacks


def _default_model_callback(model: BaseModel, **kwargs) -> BaseModel:
    # we could filter kwargs, but it's better not to - this will detect if we pass useless stuff
    return model.prepare(**kwargs)


@d.dedent
def composition(
    adata: AnnData,
    key: str,
    fontsize: Optional[str] = None,
    figsize: Optional[Tuple[float, float]] = None,
    dpi: Optional[float] = None,
    save: Optional[Union[str, pathlib.Path]] = None,
    **kwargs: Any,
) -> None:
    """Plot a pie chart for categorical annotation.

    Parameters
    ----------
    %(adata)s
    key
        Key in :attr:`~anndata.AnnData.obs` containing categorical observation.
    fontsize
        Font size for the pie chart labels.
    %(plotting)s
    kwargs
        Keyword arguments for :meth:`~matplotlib.axes.Axes.pie`.

    Returns
    -------
    %(just_plots)s
    """
    if key not in adata.obs:
        raise KeyError(f"Data not found in `adata.obs[{key!r}]`.")
    if not isinstance(adata.obs[key].dtype, pd.CategoricalDtype):
        raise TypeError(f"Expected `adata.obs[{key!r}]` is not `categorical`, found `{infer_dtype(adata.obs[key])}`.")

    colors = adata.uns.get(f"{key}_colors", None)
    x = adata.obs[key].value_counts()

    # plot these fractions in a pie plot
    fig, ax = plt.subplots(figsize=figsize, dpi=dpi)

    ax.pie(
        x=x,
        labels=x.index,
        colors=colors,
        textprops={"fontsize": fontsize},
        **kwargs,
    )
    ax.set_title(f"composition by {key}")

    if save is not None:
        save_fig(fig, save)


# modified from: https://github.com/CarlEkerot/held-karp
def _held_karp(dists: np.ndarray) -> Tuple[float, np.ndarray]:
    """Held-Karp algorithm solves the Traveling Salesman Problem.

    This algorithm uses dynamic programming with memoization.

    Parameters
    ----------
    dists
        Distance matrix.

    Returns
    -------
    The cost and the path.
    """
    n = len(dists)

    # Maps each subset of the nodes to the cost to reach that subset, as well
    # as what node it passed before reaching this subset.
    # Node subsets are represented as set bits.
    C = {}

    # Set transition cost from initial state
    for k in range(1, n):
        C[1 << k, k] = (dists[0][k], 0)

    # Iterate subsets of increasing length and store intermediate results
    # in classic dynamic programming manner
    for subset_size in range(2, n):
        for subset in itertools.combinations(range(1, n), subset_size):
            # Set bits for all nodes in this subset
            bits = 0
            for bit in subset:
                bits |= 1 << bit

            # Find the lowest cost to get to this subset
            for k in subset:
                prev = bits & ~(1 << k)

                res = []
                for m in subset:
                    if m == 0 or m == k:
                        continue
                    res.append((C[prev, m][0] + dists[m][k], m))
                C[bits, k] = min(res)

    # We're interested in all bits but the least significant (the start state)
    bits = (2**n - 1) - 1

    # Calculate optimal cost
    res = []
    for k in range(1, n):
        res.append((C[bits, k][0] + dists[k][0], k))
    opt, parent = min(res)

    # Backtrack to find full path
    path = []
    for _ in range(n - 1):
        path.append(parent)
        new_bits = bits & ~(1 << parent)
        _, parent = C[bits, parent]
        bits = new_bits

    # Add implicit start state
    path.append(0)

    return opt, np.array(path)[::-1]


def _get_categorical_colors(adata: AnnData, cluster_key: str) -> Tuple[np.ndarray, Mapping[str, str]]:
    if cluster_key not in adata.obs:
        raise KeyError(f"Unable to find data in `adata.obs[{cluster_key!r}].`")
    if not isinstance(adata.obs[cluster_key].dtype, pd.CategoricalDtype):
        raise TypeError(
            f"Expected `adata.obs[{cluster_key!r}]` to be categorical, "
            f"found `{infer_dtype(adata.obs[cluster_key])}`."
        )

    color_key = f"{cluster_key}_colors"
    try:
        colors = adata.uns[color_key]
    except KeyError:
        adata.uns[color_key] = colors = _create_categorical_colors(len(adata.obs[cluster_key].cat.categories))
    mapper = dict(zip(adata.obs[cluster_key].cat.categories, colors))
    mapper[np.nan] = "grey"

    return colors, mapper


def _get_sorted_colors(
    adata: AnnData,
    cluster_key: Union[str, Sequence[str]],
    time_key: Optional[str] = None,
    tmin: float = -np.inf,
    tmax: float = np.inf,
) -> List[np.ndarray]:
    if time_key is not None:
        if time_key not in adata.obs:
            raise KeyError(f"Unable to find time in `adata.obs[{time_key!r}]`.")
        adata = adata[(adata.obs[time_key] >= tmin) & (adata.obs[time_key] <= tmax)]
        if not adata.n_obs:
            raise ValueError(f"Specified time range `{[tmin, tmax]}` does not contain any data.")
        order = np.argsort(adata.obs[time_key].values)
    else:
        order = np.arange(adata.n_obs)

    if isinstance(cluster_key, str):
        cluster_key = (cluster_key,)
    cluster_key = _unique_order_preserving(cluster_key)

    res = []
    for ck in cluster_key:
        try:
            cols, mapper = _get_categorical_colors(adata, ck)
            res.append(np.array([colors.to_hex(mapper[v]) for v in adata.obs[ck].values[order]]))
        except TypeError:
            if not np.issubdtype(adata.obs[ck].dtype, np.number):
                raise TypeError(
                    f"Expected `adata.obs[{cluster_key!r}]` to be numeric, "
                    f"found `{infer_dtype(adata.obs[cluster_key])}`."
                ) from None
            res.append(np.asarray(adata.obs[ck])[order])

    return res



import enum
import os
import pathlib
from typing import Any, Literal, Tuple, Union

from anndata import AnnData
from scanpy import read

from cellrank import logging as logg
from cellrank._utils._docs import d, inject_docs
from cellrank._utils._enum import ModeEnum

__all__ = [
    "pancreas",
    "lung",
    "reprogramming_morris",
    "reprogramming_schiebinger",
    "zebrafish",
    "bone_marrow",
]


class ReprogrammingSubset(ModeEnum):
    FULL = enum.auto()
    K48 = "48k"
    K85 = "85k"


# fmt: off
_datasets = {
    "pancreas": ("https://figshare.com/ndownloader/files/25060877", (2531, 27998)),
    "pancreas_preprocessed": ("https://figshare.com/ndownloader/files/25030028", (2531, 2000)),
    "pancreas_preprocessed_vk": ("https://figshare.com/ndownloader/files/41325411", (2531, 5974)),
    "lung": ("https://figshare.com/ndownloader/files/25038224", (24882, 24051)),
    "reprogramming_morris": ("https://figshare.com/ndownloader/files/25503773", (104679, 22630)),
    "zebrafish": ("https://figshare.com/ndownloader/files/27265280", (2434, 23974)),
    "reprogramming_schiebinger": ("https://figshare.com/ndownloader/files/28618734", (236285, 19089)),
    "reprogramming_schiebinger_serum_subset": ("https://figshare.com/ndownloader/files/35858033", (165892, 19089)),
    "bone_marrow": ("https://figshare.com/ndownloader/files/35826944", (5780, 27876)),
}
# fmt: on


def _load_dataset_from_url(
    fpath: Union[str, pathlib.Path], url: str, expected_shape: Tuple[int, int], **kwargs: Any
) -> AnnData:
    fpath = str(fpath)
    if not fpath.endswith(".h5ad"):
        fpath += ".h5ad"

    if os.path.isfile(fpath):
        logg.debug(f"Loading dataset from `{fpath!r}`")
    else:
        logg.debug(f"Downloading dataset from `{url!r}` as `{fpath!r}`")

    dirname, _ = os.path.split(fpath)
    try:
        if not os.path.isdir(dirname):
            logg.debug(f"Creating directory `{dirname!r}`")
            os.makedirs(dirname, exist_ok=True)
    except OSError as e:
        logg.debug(f"Unable to create directory `{dirname!r}`. Reason `{e}`")

    kwargs.setdefault("sparse", True)
    kwargs.setdefault("cache", True)

    adata = read(fpath, backup_url=url, **kwargs)

    if adata.shape != expected_shape:
        raise ValueError(f"Expected `anndata.AnnData` object to have shape `{expected_shape}`, found `{adata.shape}`.")

    adata.var_names_make_unique()

    return adata


def pancreas(
    path: Union[str, pathlib.Path] = "datasets/endocrinogenesis_day15.5.h5ad",
    kind: Literal["raw", "preprocessed", "preprocessed-kernel"] = "raw",
    **kwargs: Any,
) -> AnnData:  # pragma: no cover
    """Development of the murine pancreas at E15.5 from :cite:`bastidas-ponce:19`.

    scRNA-seq dataset containing 2531 cells recorded using 10x Chromium in a single time point.
    Data was filtered to remove heavily cycling populations and to focus on the late stages of endocrinogenesis.

    Parameters
    ----------
    path
        Path where to save the dataset.
    kind
        What kind of dataset to download. Valid option are:

        - ``'raw'`` - contains raw spliced and unspliced count data, low-dimensional embedding coordinates
          as well as original cluster annotations.
        - ``'preprocessed'`` - same as above, but additionally genes have been filtered to contain at least
          20 unspliced and spliced counts and subsetted to the top 2000 highly variable genes. Data has been
          normalized by total counts and log-transformed.
        - ``'preprocessed-kernel'`` - same as above, but additionally a cell-cell transition matrix has been computed
          using the :class:`~cellrank.kernels.VelocityKernel`.
    kwargs
        Keyword arguments for :func:`~scanpy.read`.

    Returns
    -------
    Annotated data object.
    """
    path, ext = os.path.splitext(path)
    path = f"{path}_{kind}{ext}"

    if kind == "raw":
        return _load_dataset_from_url(path, *_datasets["pancreas"], **kwargs)
    if kind == "preprocessed":
        return _load_dataset_from_url(path, *_datasets["pancreas_preprocessed"], **kwargs)
    if kind == "preprocessed-kernel":
        return _load_dataset_from_url(path, *_datasets["pancreas_preprocessed_vk"], **kwargs)
    raise ValueError(f"Unknown dataset kind `{kind!r}`.")


def lung(
    path: Union[str, pathlib.Path] = "datasets/lung_regeneration.h5ad",
    **kwargs: Any,
) -> AnnData:  # pragma: no cover
    """Regeneration of murine lung epithelial cells at 13 time points from :cite:`strunz:20`.

    sc-RNA-seq dataset comprising `24 051` cells recorded using Drop-seq :cite:`macosko:15` at 13 time points spanning
    days 2-15 past lung bleomycin injury. Data was filtered to remove control cells as well as later time points which
    are more spaced out. We wanted to focus on the densely sampled days when RNA velocity :cite:`manno:18,bergen:20`
    can be used to predict the future cellular state.

    Contains raw spliced and unspliced count data, low-dimensional embedding coordinates as well as original
    cluster annotations.

    Parameters
    ----------
    path
        Path where to save the dataset.
    kwargs
        Keyword arguments for :func:`~scanpy.read`.

    Returns
    -------
    Annotated data object.
    """
    return _load_dataset_from_url(path, *_datasets["lung"], **kwargs)


@inject_docs(s=ReprogrammingSubset)
@d.dedent
def reprogramming_morris(
    path: Union[str, pathlib.Path] = "datasets/reprogramming_morris.h5ad",
    subset: Literal["full", "48k", "85k"] = ReprogrammingSubset.FULL,
    **kwargs: Any,
) -> AnnData:  # pragma: no cover
    """Reprogramming of mouse embryonic fibroblasts to induced endoderm progenitors from :cite:`morris:18`.

    sc-RNA-seq dataset comprising `104 887` cell recorded using 10X Chromium and Drop-seq :cite:`macosko:15`
    at 8 time points spanning days 0-28 past reprogramming initiation.

    Contains raw spliced and unspliced count data, low-dimensional embedding coordinates as well as clonal information
    from CellTagging :cite:`morris:18`. Moreover, it contains the following :attr:`~anndata.AnnData.obs` annotations:

    - ``'reprogramming_day'`` - time-point information.
    - ``'reprogramming'`` - whether this clone is enriched for cells from successfully reprogrammed populations.
    - ``'CellTagDN_XXk'`` - CellTag from day `N` from the `XXk` cells ``subset``.

    Parameters
    ----------
    path
        Path where to save the dataset.
    subset
        Whether to return the full object or just a subset. Can be one of:

        - ``{s.FULL!r}`` - return the complete dataset containing `104 887` cells.
        - ``{s.K85!r}`` - return the subset as described in :cite:`morris:18` Fig. 1, containing `85 010` cells.
        - ``{s.K48!r}`` - return the subset as described in :cite:`morris:18` Fig. 3, containing `48 515` cells.
    kwargs
        Keyword arguments for :func:`~scanpy.read`.

    Returns
    -------
    Annotated data object.

    Notes
    -----
    The dataset has approximately 1.5GiB and the subsetting is performed locally after the full download.
    """
    subset = ReprogrammingSubset(subset)
    adata = _load_dataset_from_url(path, *_datasets["reprogramming_morris"], **kwargs)

    if subset == ReprogrammingSubset.FULL:
        return adata
    if subset == ReprogrammingSubset.K48:
        return adata[~adata.obs["cluster"].isnull()].copy()
    if subset == ReprogrammingSubset.K85:
        return adata[~adata.obs["timecourse"].isnull()].copy()

    raise NotImplementedError(f"Subsetting option `{subset!r}` is not yet implemented.")


@d.dedent
def reprogramming_schiebinger(
    path: Union[str, pathlib.Path] = "datasets/reprogramming_schiebinger.h5ad",
    subset_to_serum: bool = False,
    **kwargs: Any,
) -> AnnData:  # pragma: no cover
    """Reprogramming of mouse embryonic fibroblasts to induced pluripotent stem cells from :cite:`schiebinger:19`.

    sc-RNA-seq dataset comprising `236 285` cell recorded using 10X Chromium
    at 39 time points spanning days 0-18 past reprogramming initiation.

    Contains total-counts normalized, log-transformed counts and low-dimensional embedding coordinates (force-directed).
    Moreover, it contains the following :attr:`~anndata.AnnData.obs` annotations:

    - ``'day'`` - time-point information.
    - ``'serum'``/``'2i'`` - whether this cell comes from the serum/2i condition.
    - ``'cell_sets'`` - cluster labels.

    Parameters
    ----------
    path
        Path where to save the dataset.
    subset_to_serum
        Whether to return the full object or subsetted to the serum condition.
        This subset also contains the pre-computed transition matrix.
    kwargs
        Keyword arguments for :func:`~scanpy.read`.

    Returns
    -------
    Annotated data object.

    Notes
    -----
    The full dataset has approximately 1.4GiB.
    """
    if subset_to_serum:
        key = "reprogramming_schiebinger_serum_subset"
        path, ext = os.path.splitext(path)
        path = f"{path}_serum{ext}"
    else:
        key = "reprogramming_schiebinger"

    return _load_dataset_from_url(path, *_datasets[key], **kwargs)


@d.dedent
def zebrafish(
    path: Union[str, pathlib.Path] = "datasets/zebrafish.h5ad",
    **kwargs: Any,
) -> AnnData:  # pragma: no cover
    """Zebrafish embryogenesis assayed using drop-seq from :cite:`farrell:18`.

    sc-RNA-seq time-series dataset, restricted to the axial mesoderm lineage, comprising of `2434` cells which
    contains 12 time-points spanning 3.3-12 hours past fertilization.

    Parameters
    ----------
    path
        Path where to save the dataset.
    kwargs
        Keyword arguments for :func:`~scanpy.read`.

    Returns
    -------
    Annotated data object.
    """
    return _load_dataset_from_url(path, *_datasets["zebrafish"], **kwargs)


@d.dedent
def bone_marrow(
    path: Union[str, pathlib.Path] = "datasets/bone_marrow.h5ad",
    **kwargs: Any,
) -> AnnData:  # pragma: no cover
    """sc-RNA-seq dataset early human hematopoiesis (CD34+ bone marrow cells) assayed using 10X Chromium.

    This dataset contains raw spliced and unspliced counts estimated using *velocyto* :cite:`manno:18`.
    Furthermore, the dataset also contains a precomputed *Palantir* pseudotime :cite:`setty:19`.

    Parameters
    ----------
    path
        Path where to save the dataset.
    kwargs
        Keyword arguments for :func:`~scanpy.read`.

    Returns
    -------
    Annotated data object.
    """
    return _load_dataset_from_url(path, *_datasets["bone_marrow"], **kwargs)
